import abc
from collections.abc import Hashable
from typing import Any

from gen_epix.fastapp import model
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


class BaseUserManager(abc.ABC):
    """
    Class that defines the interface for a user manager. This class should be
    subclassed and the methods implemented to provide the necessary functionality for
    the user manager to work with the system. The user manager is responsible for
    creating, retrieving, and managing users in the system.

    The claims passed to the methods in this class are e.g. the claims extracted from
    a JWT token.
    """

    def get_user_key_from_claims(self, claims: dict[str, Any]) -> str | None:
        """
        Get the user key, which uniquely identifies the user across systems, from the
        claims. The email claim is used here as the default user key, override this
        method if another key should be used.
        """
        email: str = claims.get("email").lower() if "email" in claims else None
        return email.lower() if email else None

    @abc.abstractmethod
    def get_user_instance_from_claims(
        self, claims: dict[str, Any]
    ) -> model.User | None:
        raise NotImplementedError

    @abc.abstractmethod
    def create_root_user_from_claims(self, claims: dict[str, Any]) -> model.User:
        raise NotImplementedError

    @abc.abstractmethod
    def is_root_user_claims(self, claims: dict[str, Any]) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def is_root_user(self, user: model.User) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def create_user_from_claims(self, claims: dict[str, Any]) -> model.User | None:
        raise NotImplementedError

    @abc.abstractmethod
    def create_new_user_from_token(
        self, user: model.User, token: str, **kwargs: Any
    ) -> model.User:
        raise NotImplementedError

    @abc.abstractmethod
    def is_existing_user_by_key(
        self, user_key: str | None, uow: BaseUnitOfWork
    ) -> bool:
        """
        Check if a user exists by their key.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        """
        Retrieve an existing user by their key.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_by_id(self, user_id: Hashable) -> model.User:
        """
        Retrieve an existing user by their id.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_permissions(self, user: model.User) -> set[model.Permission]:
        """
        Retrieve the permissions for a user instance.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_user_name_from_claims(self, claims: dict[str, Any]) -> str | None:
        """
        Extract a user-friendly display name from raw claims (dict),
        mirroring the AuthService.get_name_from_claims priority and handling
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def update_user_name(self, user: model.User, new_name: str) -> model.User | None:
        """
        Update the user's name in the user manager.
        This method should be implemented to update the user's name in the user manager.
        """
        raise NotImplementedError()



from typing import Any

from gen_epix.casedb.domain import command, exc, model
from gen_epix.casedb.domain.service import BaseAbacService
from gen_epix.fastapp.model import Policy


class BaseCaseAbacPolicy(Policy):
    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        self.abac_service = abac_service
        self.props = kwargs

    @staticmethod
    def get_case_abac_from_command(
        cmd: command.Command,
    ) -> model.CaseAbac | None:
        case_abac: model.CaseAbac | None = None
        for policy in cmd._policies:
            if not issubclass(type(policy), BaseCaseAbacPolicy):
                continue
            if case_abac:
                raise exc.InitializationServiceError(
                    f"Multiple policies registered to retrieve CaseAbac"
                )
            case_abac = policy.get_content(cmd)

        return case_abac



# pylint: disable=useless-import-alias
from gen_epix.seqdb.domain.model.seq.base import AlignmentMixin as AlignmentMixin
from gen_epix.seqdb.domain.model.seq.base import CodeMixin as CodeMixin
from gen_epix.seqdb.domain.model.seq.base import ProtocolMixin as ProtocolMixin
from gen_epix.seqdb.domain.model.seq.base import QualityMixin as QualityMixin
from gen_epix.seqdb.domain.model.seq.base import SeqMixin as SeqMixin
from gen_epix.seqdb.domain.model.seq.metadata import (
    AlignmentProtocol as AlignmentProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import (
    AssemblyProtocol as AssemblyProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import AstProtocol as AstProtocol
from gen_epix.seqdb.domain.model.seq.metadata import (
    KmerDetectionProtocol as KmerDetectionProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import (
    LibraryPrepProtocol as LibraryPrepProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import Locus as Locus
from gen_epix.seqdb.domain.model.seq.metadata import (
    LocusDetectionProtocol as LocusDetectionProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import LocusSet as LocusSet
from gen_epix.seqdb.domain.model.seq.metadata import LocusSetMember as LocusSetMember
from gen_epix.seqdb.domain.model.seq.metadata import PcrProtocol as PcrProtocol
from gen_epix.seqdb.domain.model.seq.metadata import RefAllele as RefAllele
from gen_epix.seqdb.domain.model.seq.metadata import RefSnp as RefSnp
from gen_epix.seqdb.domain.model.seq.metadata import RefSnpSet as RefSnpSet
from gen_epix.seqdb.domain.model.seq.metadata import RefSnpSetMember as RefSnpSetMember
from gen_epix.seqdb.domain.model.seq.metadata import SeqCategory as SeqCategory
from gen_epix.seqdb.domain.model.seq.metadata import SeqCategorySet as SeqCategorySet
from gen_epix.seqdb.domain.model.seq.metadata import (
    SeqClassificationProtocol as SeqClassificationProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import (
    SnpDetectionProtocol as SnpDetectionProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import SubtypingScheme as SubtypingScheme
from gen_epix.seqdb.domain.model.seq.metadata import Taxon as Taxon
from gen_epix.seqdb.domain.model.seq.metadata import TaxonLocusLink as TaxonLocusLink
from gen_epix.seqdb.domain.model.seq.metadata import (
    TaxonomyProtocol as TaxonomyProtocol,
)
from gen_epix.seqdb.domain.model.seq.metadata import TaxonSet as TaxonSet
from gen_epix.seqdb.domain.model.seq.metadata import TaxonSetMember as TaxonSetMember
from gen_epix.seqdb.domain.model.seq.metadata import TreeAlgorithm as TreeAlgorithm
from gen_epix.seqdb.domain.model.seq.metadata import (
    TreeAlgorithmClass as TreeAlgorithmClass,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    CompleteAlleleProfile as CompleteAlleleProfile,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    CompleteContig as CompleteContig,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    CompleteSample as CompleteSample,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import CompleteSeq as CompleteSeq
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    CompleteSnpProfile as CompleteSnpProfile,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    MultipleAlignment as MultipleAlignment,
)
from gen_epix.seqdb.domain.model.seq.non_persistable import (
    PhylogeneticTree as PhylogeneticTree,
)
from gen_epix.seqdb.domain.model.seq.persistable import Allele as Allele
from gen_epix.seqdb.domain.model.seq.persistable import (
    AlleleAlignment as AlleleAlignment,
)
from gen_epix.seqdb.domain.model.seq.persistable import AlleleProfile as AlleleProfile
from gen_epix.seqdb.domain.model.seq.persistable import AstMeasurement as AstMeasurement
from gen_epix.seqdb.domain.model.seq.persistable import AstPrediction as AstPrediction
from gen_epix.seqdb.domain.model.seq.persistable import (
    ContigAlignment as ContigAlignment,
)
from gen_epix.seqdb.domain.model.seq.persistable import KmerProfile as KmerProfile
from gen_epix.seqdb.domain.model.seq.persistable import PcrMeasurement as PcrMeasurement
from gen_epix.seqdb.domain.model.seq.persistable import RawSeq as RawSeq
from gen_epix.seqdb.domain.model.seq.persistable import ReadSet as ReadSet
from gen_epix.seqdb.domain.model.seq.persistable import RefSeq as RefSeq
from gen_epix.seqdb.domain.model.seq.persistable import Sample as Sample
from gen_epix.seqdb.domain.model.seq.persistable import Seq as Seq
from gen_epix.seqdb.domain.model.seq.persistable import SeqAlignment as SeqAlignment
from gen_epix.seqdb.domain.model.seq.persistable import SeqCategory as SeqCategory
from gen_epix.seqdb.domain.model.seq.persistable import (
    SeqClassification as SeqClassification,
)
from gen_epix.seqdb.domain.model.seq.persistable import SeqDistance as SeqDistance
from gen_epix.seqdb.domain.model.seq.persistable import (
    SeqDistanceProtocol as SeqDistanceProtocol,
)
from gen_epix.seqdb.domain.model.seq.persistable import SeqMixin as SeqMixin
from gen_epix.seqdb.domain.model.seq.persistable import SeqTaxonomy as SeqTaxonomy
from gen_epix.seqdb.domain.model.seq.persistable import SnpProfile as SnpProfile



# pylint: disable=wildcard-import, unused-wildcard-import
# because this is a package, and imported as such in other modules

from gen_epix.fastapp.api.exc import *
from gen_epix.fastapp.exc import *



import random
import string
import time
import uuid
from abc import ABC
from collections import Counter, defaultdict
from collections.abc import Collection, Iterable
from datetime import date, datetime, timedelta
from pathlib import Path
from typing import Any

import sqlalchemy as sa

from gen_epix.fastapp import BaseRepository, Domain
from gen_epix.omopdb.domain.enum import AnonMethod, AnonStrictness, ServiceType
from gen_epix.omopdb.domain.model.base import Model


class BaseAnonymizer(ABC):
    """
    Functionality that is identical for all Anonymizers, regardless of the type of object they anonymize
    Primarily has datatype / anonymization-method specific functionalities
    Handles anonymization methods & methods: random, categorical, shift
        random: str, int, uuid
        categorical: handles all
        shift: datetime, date

    N.B. method make_null is contemplated but not implemented, if other data types of needed, add them here

    """

    # Some reasonable defaults than can be overidden with keyword arguments
    _MAX_DAYS_OFFSET = 365
    _MIN_DAYS_OFFSET = 1
    _MIN_VALUES_REQUIRED = 10  # for categorical replacements, require a value occurs this many times to be a candidate for use
    _ALLOW_FUTURE_DATES = False

    def __init__(self, seed: int | None = None, **kwargs: Any) -> None:

        self.min_days_offset: int = (
            kwargs.get("min_days_offset") or self._MIN_DAYS_OFFSET
        )
        self.max_days_offset: int = (
            kwargs.get("max_days_offset") or self._MAX_DAYS_OFFSET
        )
        self.min_values_required: int = (
            kwargs.get("min_values_required") or self._MIN_VALUES_REQUIRED
        )
        self.allow_future_dates: bool = (
            kwargs.get("allow_future_dates") or self._ALLOW_FUTURE_DATES
        )

        if not seed:
            seed = int(time.time())
        elif not isinstance(seed, int):
            raise ValueError("Seed must be an integer")
        random.seed(seed)

    def anonymize_dates(self, datelike_values: Collection[date]) -> dict[date, date]:
        """
        Takes a collection of dates and datetimes (may be mixed) and returns a dict that maps these values
        to new date / datetime values. The new dates are such that the new dates preserve the order of dates
        and the time differences in the original such that the new dates preserve the order of dates and the
        time differences in the original collection but has an arbitrary offset from the first date. Offsets
        are integer dates regardless of the data type. No distinction is made between dates and datetimes
        in the treatment, but the output will be of the same type as the input based on python's datetime.datetime
        being a subclass of datetime.date
        """

        datelike_values = list(set(datelike_values))
        offset_days = random.randint(
            self.min_days_offset, self.max_days_offset
        ) * random.choice([-1, 1])
        today = datetime.today().date()
        latest_date = max(
            datelike_values, key=lambda x: x.date() if isinstance(x, datetime) else x
        )
        latest_date = (
            latest_date.date() if isinstance(latest_date, datetime) else latest_date
        )
        latest_date_shifted = latest_date + timedelta(days=offset_days)
        if not self.allow_future_dates and latest_date_shifted > today:
            offset_days = offset_days - (latest_date_shifted - today).days
        offset_days_as_timedelta = timedelta(days=offset_days)
        mapped = {d: d + offset_days_as_timedelta for d in datelike_values}
        return mapped

    def anonymize_text(self, text_values: Collection[str]) -> dict[str, str]:
        text_values = list(set(text_values))
        mapped = dict()
        for tv in text_values:
            len_text = len(tv)
            anon_text = "".join(random.choices(string.ascii_lowercase, k=len_text))
            mapped[tv] = anon_text
        return mapped

    def anonymize_ints(self, int_values: Collection[int]) -> dict[int, int]:
        """
        Take a collection of integers and returns a dict mapping each integer to a new integer
        with the same number of digits and same sign, minimum abs = 1
        """
        int_values = list(set(int_values))
        mapped = dict()
        for i in int_values:
            sign = 1 if i >= 0 else -1
            num_digits = len(str(abs(i)))
            new_int = random.randint(1, 10**num_digits - 1) * sign
            mapped[i] = new_int
        return mapped

    def anonymize_uuids(
        self, uuid_values: Collection[uuid.UUID]
    ) -> dict[uuid.UUID, uuid.UUID]:
        """
        Take a collection of uuid.UUID objects and returns a dict mapping each uuid.UUID to a new uuid.UUID
        N.B. iLES has no uuids that need to be anonymized, so this has not been tested
        """
        uuid_values = list(set(uuid_values))
        mapped = {u: uuid.uuid4() for u in uuid_values}
        return mapped

    def anonymize_categorical(
        self, categorical_values: Iterable[Any], choices: Collection[Any]
    ) -> dict[Any, Any]:
        """
        Take a collection of categorical values and returns a dict mapping each value to a new value
        among the choices provided. Excludes None and "" from the choices, as that may cause conflicts even
        if it's allowable in the data or just be confusing to the user.

        TODO (possible): implement proportional choices, i.e., make assignment more likely depending how
        often a value occurs in the input choices collection.

        @Ivo: By choice I've made this accept any Collection for choices, not just a set. Forcing it to
        accept a set requires the coder to take that step prior to calling, which I don't see
        the point of doing. The return is a dict, so duplicates are eliminated in this step, why
        do it elsewhere? Further, if we implement proportional choices, another the reason NOT
        to require a set as input, b/c then we'd just have to count frequency outside this method
        which is likely not relevant outside this method

        :param categorical_values: the values to be anonymized
        :param choices: the possible replacement values
        :return: a dict mapping the original values to the new values
        """
        choices = list(set(choices))
        excluded = [None, ""]
        choices = [c for c in choices if c not in excluded]
        mapped = {v: random.choice(choices) for v in categorical_values}
        return mapped


class ModelAnonymizer(BaseAnonymizer):
    """
    ModelAnonymizer handles a instances of "subject" models that are are built from models in the Domain specified.
    For instance, IlesSubject is built from the models in the Iles domain.
    A "subject" model has fields that are data values, or lists of other models. Such submodels can only contain data values
    in their fields.
    """

    def __init__(
        self,
        domain: Domain,
        repository: BaseRepository,
        seed: int | None = None,
        strictness_level: AnonStrictness = AnonStrictness.STRICT,
        **kwargs: Any,
    ) -> None:
        super().__init__(seed, **kwargs)

        self.domain = domain
        self.repository = repository
        self.strictness_level = strictness_level

        self.categorical_loc_values = self.identify_and_load_categoricals()

    def attach_model_instance(self, model: Model) -> None:
        self.model = model

    def identify_and_load_categoricals(self):
        """
        Inspect the models that are present in the domain and figure out which model/fields are specified
        as anonymization_method=AnonMethod.CATEGORICAL. For these, load all the data
        available (or NYI: possibly reduced scope) as possible substitutable values
        """
        model_classes = self.domain.get_dag_sorted_models(service_type=ServiceType.OMOP)
        categorical_locs = []
        for model_class in model_classes:
            for field_name, field in model_class.model_fields.items():
                if jse := getattr(field, "json_schema_extra", None):
                    if jse.get("anonymization_method") == AnonMethod.CATEGORICAL.value:
                        categorical_locs.append((model_class, field_name))
        # TODO: rework categorical_locs into a dict so we can group by model_class, so we only have to read each model once
        categorical_loc_values = defaultdict(lambda: defaultdict(set))
        for model_class, field_name in categorical_locs:
            mapper = self.repository.get_mapper(model_class)
            sa_model_class = (
                mapper.row_class
            )  # why is this called row_class? it's the SA model class!
            with self.repository.uow() as uow:
                # @Ivo: What there a better way through the interface you built to access a single column?
                results = uow.session.execute(
                    sa.select(sa_model_class.__table__.c[field_name])
                )
                values = [t[0] for t in results.fetchall()]
                value_counts = Counter(values)
                values_culled = set(
                    [
                        val
                        for val, ct in value_counts.items()
                        if ct >= self._MIN_VALUES_REQUIRED
                    ]
                )
                categorical_loc_values[model_class][field_name] = values_culled
        return categorical_loc_values

    def create_anonymization_map(
        self, model: Model
    ) -> dict[tuple[type, AnonMethod], list[tuple[tuple[Any], Any]]]:
        """
        Builds a map to the data to be anonymized by reading the models in the domain, which
        themselves contain an indication of whether and how each field needs to be treated.

        Output is a dict of lists:
            { (type, anon_method): [((field_name,), value), ((field_name, i), value), ...] }
            where:
                key: tuple with the datatype and method of anonymization
                value: lists of tuples, where each tuple contains a path to the field and the field value

        The first tuple in the dict value is the path to the field in the model, which can be a
        1-tuple, 2-tuple, or 3-tuple. See discussion in get_nested_attribute() regarding the limitation of
        the structures in self.model that can be handled.

        :param model: the (pydantic) model to be scanned
        """

        anon_map = defaultdict(list)
        for field_name, field in model.model_fields.items():
            value = getattr(model, field_name, None)
            type_ = type(value)
            if value is None:
                # Don't need to do anything if there's no data
                continue
            if isinstance(value, list):
                # Always work through elements of lists
                to_anonymize = True
            else:
                to_anonymize = field.json_schema_extra.get("to_anonymize", None)
                if to_anonymize:
                    anon_method = AnonMethod(
                        field.json_schema_extra.get("anonymization_method", None)
                    )
            if (value is not None or value == "") and to_anonymize is None:
                msg = (
                    f"Must declare anonymization intent for every field with data. Found "
                    f"data in {model.__class__.__name__}.{field_name} (={value}) but no anonymization intent."
                )
                if self.strictness_level == AnonStrictness.IGNORE:
                    pass
                elif self.strictness_level == AnonStrictness.WARN:
                    print("WARNING: ", msg)
                elif self.strictness_level == AnonStrictness.STRICT:
                    raise ValueError(msg)
                else:
                    raise ValueError(
                        f"No logic for strictness_level = {self.strictness_level}"
                    )
            if not to_anonymize:
                continue
            if type_ not in (Model, list):
                anon_map[type_, anon_method].append(((field_name,), value))
            elif isinstance(value, Model):
                sub_locs = self.create_anonymization_map(value)
                for typ_meth_tup, loc_value_tups in sub_locs.items():
                    anon_map[typ_meth_tup].extend(
                        [((field_name,) + k, v) for k, v in loc_value_tups]
                    )
            elif isinstance(value, list):
                for i, v in enumerate(value):
                    if isinstance(v, Model):
                        sub_locs = self.create_anonymization_map(v)
                        for typ_meth_tup, loc_value_tups in sub_locs.items():
                            anon_map[typ_meth_tup].extend(
                                [((field_name, i) + k, v) for k, v in loc_value_tups]
                            )
                    else:
                        anon_map[type_, anon_method].append(((field_name, i), v))
        return anon_map

    def anonymize(self) -> Model:
        """
        Anonymizes the model according to the specification and returns the anonymized model
        At present anonymization_spec is not handled as we haven't yet designed it.
        """

        def validate_location_path(path: tuple[str | int, ...]) -> bool:
            """
            Validate the path to ensure it is matches expected form (see get_nested_attribute)
            """
            if not isinstance(path, tuple):
                raise ValueError(f"Invalid path: {path}. Path must be a tuple")
            elif len(path) == 1 and isinstance(path[0], str):
                return True
            elif (
                len(path) == 2 and isinstance(path[0], str) and isinstance(path[1], str)
            ):
                return True
            elif (
                len(path) == 2 and isinstance(path[0], str) and isinstance(path[1], int)
            ):
                return True
            elif (
                len(path) == 3
                and isinstance(path[0], str)
                and isinstance(path[1], int)
                and isinstance(path[2], str)
            ):
                return True
            else:
                raise ValueError(
                    f"Invalid path: {path}. Path must be a 1-tuple, 2-tuple, or 3-tuple of strings and/or integers"
                )

        def get_nested_attribute(model: Model, path: tuple[str | int, ...]) -> Any:
            """
            Get the value of a nested attribute in a model using a tuple path
            that indicates where in the model the attribute is located.
            :param path: Either a 1-tuple, 2-tuple, or 3-tuple like:
                ("field_name",): specifies a scalar field at top-level
                ("field_name", "subfield_name"): specifies a scalar field within a field that is itself a model
                ("field_name", 1, "subfield_name:): specifies a scalar fields within a model that is the first instance in a list of models
                N.B. This does not handle nested models that contain lists themselves
            """
            validate_location_path(path)
            current = model
            for key in path:
                if isinstance(key, int):
                    current = current[key]
                else:
                    current = getattr(current, key)
            return current

        def get_nested_attribute_model_field_name(
            model: Model, path: tuple[str | int, ...]
        ) -> tuple[Model, str]:
            """
            Get the model and field name corresponding to the end of a path in the nested structure
            See explanation of 'path' in sister method get_nested_attribute
            """
            validate_location_path(path)
            current = model
            for i, key in enumerate(path):
                if isinstance(key, int):
                    current = current[key]
                else:
                    # If this is the last step in the path, return (model, field name)
                    if i == len(path) - 1:
                        return current, key
                    current = getattr(current, key)
            raise ValueError("Invalid path: must end in a field access.")

        def set_nested_attribute(
            model: Model, path: tuple[str | int, ...], value: Any
        ) -> None:
            """See explanation of 'path' in sister method get_nested_attribute"""
            validate_location_path(path)
            current = model
            for key in path[:-1]:
                if isinstance(key, int):
                    current = current[key]
                else:
                    current = getattr(current, key)
            setattr(current, path[-1], value)

        def map_nested_attributes(
            model: Model, path: tuple[str], mapping: dict[Any, Any]
        ) -> None:
            """See explanation of 'path' in sister method get_nested_attribute"""
            curr_value = get_nested_attribute(model, path)
            mapped_value = mapping.get(curr_value, curr_value)
            set_nested_attribute(model, path, mapped_value)

        if not hasattr(self, "model") or not isinstance(self.model, Model):
            raise ValueError("No model attached to anonymizer!")
        anonymization_map = self.create_anonymization_map(self.model)

        # Now that we've collected all the data that needs to be anonymized, let's do it!
        # Handle data types/methods each in their own way

        collected_payloads = []

        # ints/random
        # collected_ints, anonymized_int_map = {}, {}
        if loc_value_tups := anonymization_map.pop((int, AnonMethod.RANDOM), None):
            collected_ints = dict(loc_value_tups)
            anonymized_int_map = self.anonymize_ints(collected_ints.values())
            collected_payloads.append((collected_ints, anonymized_int_map))

        # strings/random
        # collected_strings, anonymized_string_map = {}, {}
        if loc_value_tups := anonymization_map.pop((str, AnonMethod.RANDOM), None):
            collected_strings = dict(loc_value_tups)
            anonymized_string_map = self.anonymize_text(collected_strings.values())
            collected_payloads.append((collected_strings, anonymized_string_map))

        # dates/shift - captures both dates and datetimes (latter subclass of former) [CHECK IF works with .pop here]
        # collected_dates, anonymized_date_map = {}, {}
        if loc_value_tups := anonymization_map.pop((datetime, AnonMethod.SHIFT), None):
            collected_dates = dict(loc_value_tups)
            anonymized_date_map = self.anonymize_dates(collected_dates.values())
            collected_payloads.append((collected_dates, anonymized_date_map))

        # categoricals
        # for these we need to reference a lookup dict of possible values distinct for each model/field
        # and make each location a separate payload item
        categorical_loc_value_tups = []
        keys = list(anonymization_map)
        for key in keys:
            if key[1] == AnonMethod.CATEGORICAL:
                categorical_loc_value_tups.extend(anonymization_map.pop(key))
        categorical_payload_items = []
        for tup in categorical_loc_value_tups:
            path, value = tup
            model_inst, field_name = get_nested_attribute_model_field_name(
                self.model, path
            )
            model_class = model_inst.__class__
            choices = self.categorical_loc_values.get(model_class, {}).get(
                field_name, None
            )
            if not choices:
                raise ValueError(
                    f"Categorical anonymization requested for {model_class}.{field_name} but no choices were found"
                )
            else:
                # Even though it's a little hokey, build the payload item to match the others
                # so we can use map_nested_attributes to process them. Hokey here means the
                # map_ is a 1:1 dict.
                # TODO: replace with general method tof choosing replacement value that could be enhanced, e.g.,
                # with a relative frequency, or excluding certain possibilities
                # map_ = {value:  random.choice(list(choices))}
                map_ = self.anonymize_categorical([value], choices)
                categorical_payload_items.append(({path: None}, map_))
        collected_payloads.extend(categorical_payload_items)

        # Check we've set up processing for all data
        if anonymization_map:
            msg = f"Could not find anomyization methods requested: {anonymization_map.keys()}"
            if self.strictness_level == AnonStrictness.IGNORE:
                pass
            elif self.strictness_level == AnonStrictness.WARN:
                print("WARNING: ", msg)
            elif self.strictness_level == AnonStrictness.STRICT:
                raise ValueError(msg)
            else:
                raise ValueError(
                    "AnonStrictness level not recognized: {self.strictness_level}"
                )
        # else:
        #    print("Found methods for all anonymization methods requested")

        # Final step: Make the substitutions in the top-level model
        anonymized_model = self.model.model_copy(deep=True)
        for location_dict, map_ in collected_payloads:
            for location in location_dict:
                map_nested_attributes(anonymized_model, location, map_)

        # Altering the model with map_nested_attributes results in dicts inside - need to convert back into model
        anonymized_model = type(anonymized_model).model_validate(
            anonymized_model.model_dump(), strict=True
        )

        self.anonymized_model = anonymized_model

        return anonymized_model

    # @John: can you move this to the unit test for the ModelAnonymizer class?
    # @Ivo: This does not fit in the unit test framework. This is meant for human inspection
    # of the anonymization to see if it's working. There's probably a better place for it
    # than here, but that can be dealt with later. Additionally, yes, unit tests could/should
    # be developed.
    def compare_models(self, fp_to_output: Path | None = None) -> None:
        """
        Prints or writes to file formatted comparison of original and anonymized models,
        for development and testing purposes.
        """
        _KEY_PAD = 60
        _VALUE_PAD = 25
        orig_model = self.model.model_dump()
        anon_model = self.anonymized_model.model_dump()
        assert orig_model.keys() == anon_model.keys()
        out_str = ""

        def assert_same_structure(a, b, path="") -> str:
            if type(a) != type(b):
                raise ValueError(
                    f"Type mismatch at {path or '<root>'}: {type(a)} vs {type(b)}"
                )

            if isinstance(a, dict):
                if a.keys() != b.keys():
                    raise ValueError(
                        f"Dict key mismatch at {path or '<root>'}: {set(a.keys())} vs {set(b.keys())}"
                    )
                for k in a:
                    assert_same_structure(a[k], b[k], f"{path}.{k}" if path else k)

            elif isinstance(a, list):
                if len(a) != len(b):
                    raise ValueError(
                        f"List length mismatch at {path or '<root>'}: {len(a)} vs {len(b)}"
                    )
                for i, (item_a, item_b) in enumerate(zip(a, b)):
                    assert_same_structure(item_a, item_b, f"{path}[{i}]")

        def comp_struct(a, b, out_str, path=""):
            if isinstance(a, dict):
                for k in a:
                    new_path = f"{path}.{k}" if path else k
                    out_str = comp_struct(a[k], b[k], out_str, new_path)

            elif isinstance(a, list):
                for i, (item_a, item_b) in enumerate(zip(a, b)):
                    new_path = f"{path}[{i}]"
                    out_str = comp_struct(item_a, item_b, out_str, new_path)

            else:
                fmt_comp = f"{path[:_KEY_PAD]:<{_KEY_PAD}} | diff = {str(a != b):<6} | {str(a)[:_VALUE_PAD]:<{_VALUE_PAD}} | {str(b)[:_VALUE_PAD]}\n"
                out_str += fmt_comp

            return out_str

        assert_same_structure(orig_model, anon_model)
        out_str = comp_struct(orig_model, anon_model, out_str)

        if fp_to_output:
            with open(fp_to_output, "w") as f:
                f.write(out_str)
        else:
            print(out_str)



from __future__ import annotations

from collections.abc import Hashable
from enum import Enum
from typing import Any

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.domain.policy.permission import NO_RBAC_PERMISSIONS
from gen_epix.commondb.domain.service import BaseRbacService
from gen_epix.fastapp import Command, Permission
from gen_epix.fastapp.app import App


class RbacService(BaseRbacService):

    def __init__(self, app: App, **kwargs: Any) -> None:
        super().__init__(app, **kwargs)
        app_impl: AppImplDetails = app.impl
        self.role_map: dict[enum.Role | Enum, str] = app_impl.role_map
        self.role_set_map: dict[enum.RoleSet | Enum, frozenset[str]] = (
            app_impl.role_set_map
        )
        self.root_role = self.role_map[enum.Role.ROOT]
        self.guest_role = self.role_map[enum.Role.GUEST]

        # Register permissions without RBAC
        for command_class, permission_type in NO_RBAC_PERMISSIONS:
            permission = self.app.domain.get_permission(command_class, permission_type)
            self.register_permission_without_rbac(permission)

    def register_policies(self) -> None:
        self.register_rbac_policies()

    def retrieve_user_roles(self, user: model.User) -> set[Hashable]:  # type: ignore[override]
        return user.roles  # type: ignore[return-value]

    def retrieve_user_is_non_rbac_authorized(self, cmd: Command) -> bool:
        """
        Check if the user is authorized to perform the command in addition to RBAC.
        A root user is always authorized, any other user must have is_active=True.
        """
        user: model.User | None = cmd.user  # type: ignore[assignment]
        if user is None:
            return False
        return user.is_active or self.root_role in user.roles

    def retrieve_user_is_root(self, user: model.User) -> bool:  # type: ignore[override]
        return self.root_role in user.roles

    def retrieve_own_permissions(
        self, cmd: command.RetrieveOwnPermissionsCommand
    ) -> set[Permission]:
        user: model.User | None = cmd.user
        if not user or not user.id:
            return set()
        return self.retrieve_user_permissions(user)

    def retrieve_sub_roles(self, cmd: command.RetrieveSubRolesCommand) -> set[str]:
        user: model.User | None = cmd.user
        if not user or not user.id or not user.roles:
            return set()
        sub_roles: set[str] = set.union(*[self.get_sub_roles(x) for x in user.roles])  # type: ignore[arg-type]
        # Special case: ROOT is included as its own sub-role
        if self.root_role in user.roles:
            sub_roles.add(self.root_role)
        return sub_roles



from gen_epix.fastapp.service import BaseService
from gen_epix.seqdb.domain.enum import ServiceType


class BaseFileService(BaseService):
    SERVICE_TYPE = ServiceType.FILE

    def register_handlers(self) -> None:
        self.register_default_crud_handlers()



import datetime
import uuid
from enum import Enum

import ulid


class TimestampFactory(Enum):
    DATETIME_NOW = lambda: datetime.datetime.now()


class IdFactory(Enum):
    UUID4 = uuid.uuid4
    ULID = lambda: ulid.api.new().uuid



from collections.abc import Callable
from typing import Any, NoReturn

from fastapi import APIRouter

from gen_epix.commondb.api.auth import create_auth_endpoints
from gen_epix.commondb.api.organization import create_organization_endpoints
from gen_epix.commondb.api.rbac import create_rbac_endpoints
from gen_epix.commondb.api.system import create_system_endpoints
from gen_epix.fastapp import App
from gen_epix.omopdb.api.omop import create_omop_endpoints
from gen_epix.omopdb.api.organization import ApiPermission
from gen_epix.omopdb.domain import enum


def create_routers(
    app: App | None = None,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    router_kwargs: dict = {},
) -> list[APIRouter]:
    assert app
    router_data = [
        # Common routers
        {
            "name": "auth",
            "create_endpoints_fn": create_auth_endpoints,
            "endpoints_function_kwargs": {"service_type": enum.ServiceType.AUTH},
        },
        {
            "name": "rbac",
            "create_endpoints_fn": create_rbac_endpoints,
            "endpoints_function_kwargs": {"service_type": enum.ServiceType.RBAC},
        },
        {
            "name": "organization",
            "create_endpoints_fn": create_organization_endpoints,
            "endpoints_function_kwargs": {
                "service_type": enum.ServiceType.ORGANIZATION,
                "api_permission_class": ApiPermission,
            },
        },
        {
            "name": "system",
            "create_endpoints_fn": create_system_endpoints,
            "endpoints_function_kwargs": {"service_type": enum.ServiceType.SYSTEM},
        },
        # Specific routers
        {
            "name": "omop",
            "create_endpoints_fn": create_omop_endpoints,
        },
    ]
    routers: list[APIRouter] = []
    for curr_router_data in router_data:
        name: str = curr_router_data["name"]  # type: ignore[assignment]
        create_endpoints_fn: Callable = curr_router_data["create_endpoints_fn"]  # type: ignore[assignment]
        router = APIRouter(tags=[name], **router_kwargs)
        endpoints_function_kwargs: dict = curr_router_data.get(  # type: ignore[assignment]
            "endpoints_function_kwargs", {}
        )
        create_endpoints_fn(
            router,
            app,
            handle_exception=handle_exception,
            **endpoints_function_kwargs,
        )
        routers.append(router)
    return routers



import abc
from collections.abc import Iterable

import gen_epix.seqdb.domain.command as seqdb_command
from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.fastapp import BaseService


class BaseSeqdbService(BaseService):
    SERVICE_TYPE = ServiceType.SEQDB

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        f(
            command.RetrievePhylogeneticTreeBySequencesCommand,
            self.retrieve_phylogenetic_tree,
        )
        f(command.RetrieveGeneticSequenceByIdCommand, self.retrieve_genetic_sequences)
        f(
            command.RetrieveGeneticSequenceFastaByIdCommand,
            self.retrieve_genetic_sequence_fasta_by_id,
        )
        f(seqdb_command.ReadSetCrudCommand, self.crud)
        f(seqdb_command.FileCrudCommand, self.crud)
        f(seqdb_command.LibraryPrepProtocolCrudCommand, self.crud)
        f(seqdb_command.AssemblyProtocolCrudCommand, self.crud)
        f(seqdb_command.SeqCrudCommand, self.crud)
        # f(command.RetrieveAlleleProfileCommand, self.retrieve_allele_profile)

    @abc.abstractmethod
    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeBySequencesCommand
    ) -> model.PhylogeneticTree | None:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_genetic_sequences(
        self,
        cmd: command.RetrieveGeneticSequenceByIdCommand,
    ) -> list[model.GeneticSequence]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_genetic_sequence_fasta_by_id(
        self,
        cmd: command.RetrieveGeneticSequenceFastaByIdCommand,
    ) -> Iterable[str]:
        raise NotImplementedError()

    # @abc.abstractmethod
    # def retrieve_allele_profile(
    #     self,
    #     cmd: command.RetrieveAlleleProfileCommand,
    # ) -> model.SeqDbAlleleProfile | list[model.SeqDbAlleleProfile]:
    #     raise NotImplementedError()



import re
import uuid
import warnings
from collections.abc import Callable, Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, Self

import sqlalchemy as sa
from sqlalchemy import Engine, delete, select
from sqlalchemy.orm import Session, sessionmaker

import gen_epix.fastapp.exc as exc
from gen_epix.fastapp import CrudOperation, Link
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.domain.link import Link
from gen_epix.fastapp.enum import CrudOperation, FieldTypeSet, IsolationLevel
from gen_epix.fastapp.model import Model
from gen_epix.fastapp.repositories.sa.engine_factory import EngineFactory
from gen_epix.fastapp.repositories.sa.mapper import BaseSAMapper, SAMapper
from gen_epix.fastapp.repositories.sa.unit_of_work import SAUnitOfWork
from gen_epix.fastapp.repository import BaseRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.filter import (
    ComparisonOperator,
    CompositeFilter,
    DateRangeFilter,
    DatetimeRangeFilter,
    EqualsBooleanFilter,
    EqualsFilter,
    EqualsNumberFilter,
    EqualsStringFilter,
    EqualsUuidFilter,
    ExistsFilter,
    Filter,
    LogicalOperator,
    NumberRangeFilter,
    NumberSetFilter,
    RangeFilter,
    StringSetFilter,
    UuidSetFilter,
)


class SARepository(BaseRepository):
    DEFAULT_MAX_INSERT_BATCH_SIZE = 2000

    @classmethod
    def create_repository(cls, **kwargs: Any) -> BaseRepository:
        entities = kwargs.pop("entities", [])
        connection_string = kwargs.pop("connection_string", None)
        file = kwargs.pop("file", None)
        if connection_string is None:
            if file is None:
                raise exc.RepositoryInitializationServiceError(
                    "Either connection_string or file must be provided"
                )
            connection_string = f"sqlite:///{Path(file).resolve().as_posix()}"
        return cls.create_sa_repository(
            entities=entities,
            connection_string=connection_string,
            **kwargs,
        )

    def __init__(self, engine: Engine, **kwargs: Any):
        register_mappers = kwargs.pop("register_mappers", True)
        # Add properties
        self._id: str = kwargs.get("id", str(uuid.uuid4()))
        self._name: str = kwargs.get("name", self._id)
        self._engine = engine

        # Create a session maker per isolation level
        self._default_isolation_level: IsolationLevel = IsolationLevel.SERIALIZABLE
        self._session_maker_by_isolation_level: dict[IsolationLevel, sessionmaker] = {
            x: sessionmaker(engine.execution_options(isolation_level=x.value))
            for x in IsolationLevel
        }

        # Initialize remaining properties
        self._mapper_by_model: dict[type[Any], BaseSAMapper] = {}
        self._mapper_by_row: dict[type[Any], BaseSAMapper] = {}
        self._uow_context_stack: list[BaseUnitOfWork] = []

        # Register mappers if necessary
        if register_mappers:
            self.register_mappers(**kwargs)

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def default_isolation_level(self) -> IsolationLevel:
        return self._default_isolation_level

    @default_isolation_level.setter
    def default_isolation_level(self, value: IsolationLevel) -> None:
        self._default_isolation_level = value

    def uow(
        self,
        **kwargs: Any,
    ) -> BaseUnitOfWork:
        if self._uow_context_stack:
            # Nested within another context -> reuse the session of that context
            if kwargs:
                raise exc.RepositoryServiceError(
                    "Cannot pass arguments when creating a nested UnitOfWork"
                )
            return SAUnitOfWork(
                self._uow_context_stack[-1].session,
                context_stack=self._uow_context_stack,
            )
        isolation_level: IsolationLevel = kwargs.pop(
            "isolation_level", self._default_isolation_level
        )
        expire_on_commit: bool = kwargs.pop("expire_on_commit", True)
        return SAUnitOfWork(
            self.get_session(
                isolation_level=isolation_level,
                expire_on_commit=expire_on_commit,
                **kwargs,
            ),
            context_stack=self._uow_context_stack,
        )

    def get_session(
        self,
        isolation_level: IsolationLevel | None = None,
        expire_on_commit: bool = False,
        **kwargs: Any,
    ) -> Session:
        isolation_level = isolation_level or self._default_isolation_level
        session: Session = self._session_maker_by_isolation_level[isolation_level](
            expire_on_commit=expire_on_commit
        )
        return session

    def register_mappers(
        self,
        entities: list[Entity] | None = None,
        field_name_map: dict[type[Model], dict[str, str]] | None = None,
        service_metadata_field_names: dict[type[Model], tuple] | None = None,
        db_metadata_field_names: dict[type[Model], tuple] | None = None,
        generate_service_metadata: (
            dict[type[Model], Callable[[Model, Hashable], dict[str, Any]]] | None
        ) = None,
        **kwargs: Any,
    ) -> None:
        """
        Default implementation to register standard mappers for a list of entities.
        """
        # Parse arguments
        entities = entities or []
        field_name_map = field_name_map or {}
        service_metadata_field_names = service_metadata_field_names or {}
        db_metadata_field_names = db_metadata_field_names or {}
        generate_service_metadata = generate_service_metadata or {}

        # Create and register mapper for each entity
        for entity in entities:
            if not entity.persistable:
                continue
            model_class = entity.model_class
            db_model_class = entity.db_model_class
            if not db_model_class:
                raise exc.RepositoryInitializationServiceError(
                    f"Entity {entity.name} has no db_model_class set"
                )
            assert issubclass(model_class, Model)
            mapper = SAMapper(
                model_class,
                db_model_class,
                field_name_map=field_name_map.get(model_class),
                service_metadata_field_names=service_metadata_field_names.get(
                    model_class
                ),
                db_metadata_field_names=db_metadata_field_names.get(model_class),
                generate_service_metadata=generate_service_metadata.get(model_class),
            )
            self.register_mapper(mapper)

    def get_mapper(self, model_class: type) -> BaseSAMapper:
        mapper = self._mapper_by_model.get(model_class)
        if not mapper:
            raise exc.RepositoryInitializationServiceError(
                f"No mapper set for Model {model_class}"
            )
        return mapper

    def register_mapper(self, mapper: BaseSAMapper) -> Self:
        for current_mapper in self._mapper_by_model.values():
            if current_mapper.row_class == mapper.row_class:
                raise exc.RepositoryInitializationServiceError(
                    f"Mapper for {current_mapper.model_class} already set"
                )
            if (
                current_mapper.schema_name == mapper.schema_name
                and current_mapper.table_name == mapper.table_name
            ):
                raise exc.RepositoryInitializationServiceError(
                    f"Mapper for {current_mapper.model_class} already set"
                )
        model_class = mapper.model_class
        self._mapper_by_model[model_class] = mapper
        return self

    def to_sql(
        self,
        user_id: Hashable | None,
        model_class: type,
        obj: Any | Iterable[Any],
        **kwargs: Any,
    ) -> Any | list[Any]:
        mapper = self._mapper_by_model[model_class]
        if isinstance(obj, model_class):
            return mapper.dump(user_id, obj, **kwargs)
        return [mapper.dump(user_id, x, **kwargs) for x in obj]

    def from_sql(
        self, model_class: type, row: Any | Iterable[Any], **kwargs: Any
    ) -> Any | list[Any]:
        mapper = self._mapper_by_model[model_class]
        if isinstance(row, Iterable):
            return [mapper.load(x, **kwargs) for x in row]
        return mapper.load(row, **kwargs)

    def crud(  # type: ignore
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        objs: Model | Iterable[Model] | None,
        obj_ids: Hashable | Iterable[Hashable] | None,
        operation: CrudOperation,
        filter: Filter | None = None,
        **kwargs,
    ) -> Model | list[Model] | Hashable | list[Hashable] | bool | list[bool] | None:
        if not isinstance(uow, SAUnitOfWork):
            raise exc.RepositoryServiceError(f"Invalid UnitOfWork: {uow}")
        session = uow.session
        BaseRepository.verify_crud_args(model_class, objs, obj_ids, operation)
        match operation:
            case CrudOperation.CREATE_ONE:
                return self.create_one(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.CREATE_SOME:
                return self.create_some(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.READ_ONE:
                return self.read_one(model_class, obj_ids, session=session, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.READ_SOME:
                return self.read_some(model_class, obj_ids, session=session, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.READ_ALL:
                return self.read_all(model_class, filter, session=session, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.UPDATE_ONE:
                return self.update_one(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.UPDATE_SOME:
                return self.update_some(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.UPSERT_ONE:
                return self.upsert_one(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.UPSERT_SOME:
                return self.upsert_some(
                    model_class, user_id, objs, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.DELETE_ONE:
                return self.delete_one(
                    model_class, user_id, obj_ids, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.DELETE_SOME:
                return self.delete_some(
                    model_class, user_id, obj_ids, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.DELETE_ALL:
                return self.delete_all(
                    model_class, user_id, filter, session=session, **kwargs  # type: ignore[arg-type]
                )
            case CrudOperation.EXISTS_ONE:
                return self.exists_one(model_class, obj_ids, session=session, **kwargs)  # type: ignore[arg-type]
            case CrudOperation.EXISTS_SOME:
                return self.exists_some(model_class, obj_ids, session=session, **kwargs)  # type: ignore[arg-type]
            case _:
                raise NotImplementedError(f"Operation {operation} not implemented")

    def create_one(
        self, model_class: type, user_id: Hashable, obj: Model, **kwargs: Any
    ) -> Model | Hashable:
        return self.create_some(model_class, user_id, [obj], **kwargs)[0]

    def create_some(
        self,
        model_class: type,
        user_id: Hashable,
        objs: Iterable[Model],
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        # Check arguments
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        return_id: bool = kwargs.get("return_id", False)  # type: ignore[assignment]
        flush = kwargs.get("flush", True)
        max_batch_size = int(
            kwargs.get("max_batch_size", self.DEFAULT_MAX_INSERT_BATCH_SIZE)
        )
        objs = objs if isinstance(objs, list) else list(objs)
        if not objs:
            return []

        # Check objs
        if not all(isinstance(x, model_class) for x in objs):
            raise ValueError("Not all objs are of the correct Model")

        # Create rows

        def _execute(session: Session) -> list[Model] | list[Hashable]:
            rows = self.to_sql(user_id, model_class, objs)
            n_rows = len(rows)
            n_batches = int(n_rows / max_batch_size) + (n_rows / max_batch_size > 0)
            if not flush and n_batches > 1:
                raise exc.RepositoryServiceError(
                    f"Creation of {n_rows} objects requires more than one (n={n_batches}) batche while flush={flush}"
                )
            for i in range(n_batches):
                slice_ = slice(
                    i * max_batch_size,
                    min((i + 1) * max_batch_size, n_rows),
                )
                rows_slice = rows[slice_]
                session.add_all(rows_slice)
                if flush:
                    session.flush()
            if return_id:
                mapper = self.get_mapper(model_class)
                get_row_id = mapper.get_row_id
                return [get_row_id(x) for x in rows]
            return self.from_sql(model_class, rows)

        created_objs = self._execute_sa(session, _execute, kwargs)
        return created_objs  # type: ignore[return-value]

    def read_one(self, model_class: type, obj_id: Hashable, **kwargs: Any) -> Model:
        return self.read_some(model_class, [obj_id], **kwargs)[0]

    def read_some(
        self, model_class: type, obj_ids: Iterable[Hashable], **kwargs: Any
    ) -> list[Model]:
        """
        :param optimize_parameter_handling, optional kwarg:
           if True, avoid parameterized query that using SQL's IN that is
           more many parameters nonperformant by creating and joining with a temporary table instead
           default = False, but possibly recommend to set dynamically based on number of parameters
        """
        # Check arguments
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        obj_ids = obj_ids if isinstance(obj_ids, list) else list(obj_ids)
        SARepository._verify_duplicate_ids(model_class, obj_ids)
        # Retrieve rows and verify result
        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        cascade_read = kwargs.get("cascade_read", False)
        optimize_parameter_handling = kwargs.get("optimize_parameter_handling", False)

        def _execute(session: Session) -> list[Model] | list[Hashable]:
            rows, row_ids = SARepository._in_session_read_some(
                mapper, session, row_class, obj_ids, optimize_parameter_handling
            )

            # Reorder objs to guarantee same order as obj_ids and at the
            # same time detect missing objs
            map_to_index = {x: i for i, x in enumerate(row_ids)}
            objs = self.from_sql(
                model_class,
                [rows[map_to_index[x]] if x in map_to_index else None for x in obj_ids],
            )
            if any(x is None for x in objs):
                invalids_ids = [x for x, y in zip(obj_ids, objs) if y is None]
                invalids_ids_str = ", ".join([str(x) for x in invalids_ids])
                raise exc.InvalidIdsError(
                    f"{model_class} object(s) do not exist: {invalids_ids_str}",
                    ids=obj_ids,
                )
            # Read links if requested and known
            # If links were not passed explicitly, retrieve them from model
            if cascade_read:
                links = kwargs.get("links", model_class.ENTITY.links)
                self._in_session_add_cascade_read(session, links, objs)

            return objs

        objs = self._execute_sa(session, _execute, kwargs)
        return objs

    def read_all(
        self, model_class: type, filter: Filter | None, **kwargs: Any
    ) -> list[Model]:
        # Check arguments
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        # Retrieve rows and generate objs
        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        get_row_id = mapper.get_row_id
        cascade_read: bool = kwargs.get("cascade_read", False)
        return_id: bool = kwargs.get("return_id", False)
        obj_filter: Filter | None = kwargs.get("obj_filter", None)

        def _execute(session: Session) -> list[Model] | list[Hashable]:
            # Get either rows or row_ids
            if return_id:
                # Select only row_ids
                stmt = select(get_row_id(row_class))
            else:
                # Select entire row
                stmt = select(row_class)
            if filter:
                # Convert filter to where clause and add to statement
                stmt = stmt.where(
                    self.get_where_clause_from_filter(row_class, mapper, filter)
                )
            if return_id:
                row_ids = [x[0] for x in session.execute(stmt).all()]
                if obj_filter:
                    # Retrieve entire rows and filter them with obj_filter, then get
                    # remaining IDs
                    stmt2 = select(row_class).where(get_row_id(row_class).in_(row_ids))
                    rows = [x[0] for x in session.execute(stmt2).all()]
                    objs = self.from_sql(model_class, rows)
                    objs = list(obj_filter.filter_rows(objs, is_model=True))
                    if len(objs) < len(row_ids):
                        row_ids = [mapper.get_id(x) for x in objs]
                objs = row_ids
            else:
                rows = [x[0] for x in session.execute(stmt).all()]
                objs = self.from_sql(model_class, rows)
                if obj_filter:
                    objs = list(obj_filter.filter_rows(objs, is_model=True))
                # Read links if needed
                if cascade_read:
                    links = kwargs.get("links", {})
                    self._in_session_add_cascade_read(session, links, objs)
            return objs

        objs = self._execute_sa(session, _execute, kwargs)
        return objs

    def update_one(
        self, model_class: type, user_id: Hashable, obj: Model, **kwargs: Any
    ) -> Model | Hashable:
        return self.update_some(model_class, user_id, [obj], **kwargs)[0]

    def update_some(
        self,
        model_class: type,
        user_id: Hashable,
        objs: Iterable[Model],
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        # Check arguments
        objs = objs if isinstance(objs, list) else list(objs)
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        flush = kwargs.get("flush", True)
        # Retrieve row
        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        row_field_names = mapper.get_row_field_names_by_set(FieldTypeSet.DATA)
        get_row_id = mapper.get_row_id
        get_metadata = mapper.generate_service_metadata

        def _execute(session: Session) -> list[Model]:
            updated_rows = self.to_sql(user_id, model_class, objs)
            obj_ids = [get_row_id(x) for x in updated_rows]
            rows, row_ids = SARepository._in_session_read_some(
                mapper, session, row_class, obj_ids
            )
            map_rows = dict(zip(row_ids, rows))
            for updated_row in updated_rows:
                is_update = False
                row = map_rows[get_row_id(updated_row)]
                # Update content
                for row_field_name in row_field_names:
                    updated_value = getattr(updated_row, row_field_name)
                    if getattr(row, row_field_name) != updated_value:
                        is_update = True
                        setattr(row, row_field_name, updated_value)
                if is_update:
                    # TODO: avoid calling get_metadata twice, once here and once as
                    # part of to_sql above. This could be done by having a property
                    # in the mapper for those metadata fields that are not generated
                    # by the database, as only those need to be set here.
                    for row_field_name, value in get_metadata(updated_row, user_id):
                        setattr(row, row_field_name, value)
            if flush:
                session.flush()
            # Generate objs
            return self.from_sql(model_class, rows)

        updated_objs = self._execute_sa(session, _execute, kwargs)
        return updated_objs

    def upsert_one(
        self, model_class: type, user_id: Hashable, obj: Model, **kwargs: Any
    ) -> Model | Hashable:
        return self.upsert_some(model_class, user_id, [obj], **kwargs)[0]

    def upsert_some(
        self,
        model_class: type,
        _user_id: Hashable,
        _objs: Iterable[Model],
        **kwargs: Any,
    ) -> list[Model] | list[Hashable]:
        raise NotImplementedError

    def delete_one(
        self, model_class: type, user_id: Hashable, row_id: Hashable, **kwargs: Any
    ) -> Hashable:
        return self.delete_some(model_class, user_id, [row_id], **kwargs)[0]

    def delete_some(
        self,
        model_class: type,
        user_id: Hashable,
        row_ids: Iterable[Hashable],
        **kwargs: Any,
    ) -> list[Hashable]:
        # Check arguments
        row_ids = row_ids if isinstance(row_ids, list) else list(row_ids)
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        flush = kwargs.get("flush", True)
        # Delete rows
        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        get_row_id = mapper.get_row_id

        def _execute(session: Session) -> None:
            is_existing = self.exists_some(model_class, row_ids)
            if not all(is_existing):
                invalid_ids = [x for x, y in zip(row_ids, is_existing) if not y]
                invalid_ids_str = ", ".join([str(x) for x in invalid_ids])
                raise exc.InvalidIdsError(
                    f"{model_class} object(s) do not exist: {invalid_ids_str}",
                    ids=invalid_ids,
                )
            session.execute(delete(row_class).where(get_row_id(row_class).in_(row_ids)))
            if flush:
                session.flush()

        self._execute_sa(session, _execute, kwargs)
        return row_ids

    def delete_all(
        self,
        model_class: type,
        user_id: Hashable,
        filter: Filter | None,
        **kwargs: Any,
    ) -> list[Hashable] | None:
        # Check arguments
        session: Session = kwargs.get("session")  # type: ignore[assignment]
        # Delete rows
        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        get_row_id = mapper.get_row_id
        return_id: bool = kwargs.get("return_id", False)  # type: ignore[assignment]
        obj_filter: Filter | None = kwargs.get("obj_filter", None)

        def _execute(session: Session) -> list[Hashable] | None:

            if filter or obj_filter:
                raise NotImplementedError("Filter not implemented for delete_all")
            row_ids = None
            if return_id:
                # Get ids
                row_ids = session.execute(select(get_row_id(row_class))).all()
                row_ids = [x[0] for x in row_ids]

            # Delete all rows
            # TODO: commented out for sqlite workourand. Restore ASAP.
            # session.execute(delete(row_class))

            # TODO: workaround for SQLite foreign key constraint issues. Remove ASAP.
            # Check if this is SQLite and if we need to handle foreign key constraints
            is_sqlite = "sqlite" in str(session.get_bind().url).lower()
            needs_fk_workaround = False
            # For SQLite with schemas (using ATTACH DATABASE), foreign key constraints
            # can cause issues when referencing tables across schemas. Check if this table
            # has foreign keys that might need special handling.
            if is_sqlite and hasattr(row_class, "__table__"):
                table = row_class.__table__
                for fk in table.foreign_keys:
                    # If foreign key references a table in the same schema, it should work
                    # If it references a different schema or has schema prefix issues, we might need workaround
                    referenced_table = fk.column.table
                    if referenced_table.schema != table.schema or table.name in [
                        "measurement_relation"
                    ]:  # Known problematic tables
                        needs_fk_workaround = True
                        break
            # Temporarily disable foreign key constraints if needed
            if needs_fk_workaround:
                session.execute(sa.text("PRAGMA foreign_keys=OFF"))
                session.flush()
            # Delete rows
            session.execute(delete(row_class))
            # Re-enable foreign keys if we disabled them
            if needs_fk_workaround:
                session.execute(sa.text("PRAGMA foreign_keys=ON"))
                session.flush()
            # TODO: end of workaround for SQLite foreign key constraint issues. Remove ASAP.

            return row_ids

        deleted_row_ids = self._execute_sa(session, _execute, kwargs)
        return deleted_row_ids if return_id else None

    def exists_one(self, model_class: type, obj_id: Hashable, **kwargs: Any) -> bool:
        return self.exists_some(model_class, [obj_id], **kwargs)[0]

    def exists_some(
        self, model_class: type, obj_ids: Iterable[Hashable], **kwargs: Any
    ) -> list[bool]:
        session: Session = kwargs.get("session")  # type: ignore[assignment]

        mapper = self.get_mapper(model_class)
        row_class = mapper.row_class
        SARepository._verify_duplicate_ids(model_class, obj_ids)

        def _execute(session: Session) -> list[bool]:
            # select(mapper.get_row_id(row_class)) works because mapper.get_row_id returns the attribute of the row class
            # that functions as the primary key and this attribute is an SQLalchemy Column object that is aware of which table it is in
            rows: Sequence = session.execute(
                select(mapper.get_row_id(row_class)).where(
                    mapper.get_row_id(row_class).in_(obj_ids)
                )
            ).all()
            found_obj_ids = {x[0] for x in rows}
            is_existing_obj = [x in found_obj_ids for x in obj_ids]
            return is_existing_obj

        return self._execute_sa(session, _execute, kwargs)

    def read_fields(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: type[Model],
        field_names: list[str],
        filter: Filter | None = None,
        **kwargs: Any,
    ) -> Iterable[tuple]:
        if not isinstance(uow, SAUnitOfWork):
            raise exc.RepositoryServiceError(f"Invalid UnitOfWork: {uow}")
        mapper = self.get_mapper(model_class)
        field_name_map = mapper.get_field_name_map()
        row_field_names = [field_name_map[x] for x in field_names]
        row_class = mapper.row_class

        def _execute(session: Session) -> Iterable[tuple]:
            stmt = select(*[getattr(row_class, x) for x in row_field_names])
            if filter:
                # Convert filter to where clause and add to statement
                stmt = stmt.where(
                    self.get_where_clause_from_filter(row_class, mapper, filter)
                )
            for row in session.execute(stmt):
                yield row

        return self._execute_sa(uow.session, _execute, kwargs)

    def split_filter(
        self, model_class: type, filter: Filter | None
    ) -> tuple[Filter | None, Filter | None]:
        if not filter:
            return None, None
        field_name_map = self.get_mapper(model_class).get_field_name_map()
        return self._split_filter_recursion(field_name_map, filter)

    def get_where_clause_from_filter(
        self, row_class: type, mapper: BaseSAMapper, filter: Filter
    ) -> Any:
        invert = filter.invert
        if isinstance(filter, CompositeFilter):
            args = []
            for sub_filter in filter.filters:
                args.append(
                    self.get_where_clause_from_filter(row_class, mapper, sub_filter)
                )
            if filter.operator == LogicalOperator.AND:
                return sa.and_(*args) if not invert else sa.not_(sa.and_(*args))
            if filter.operator == LogicalOperator.OR:
                return sa.or_(*args) if not invert else sa.not_(sa.or_(*args))
            raise exc.InvalidArgumentsError(
                f"Unsupported filter operator: {filter.operator.value}"
            )
        row_field_name = mapper.get_mapped_field_name(str(filter.get_key()))
        if row_field_name is None:
            raise exc.InvalidArgumentsError(
                f"Filter key '{filter.get_key()}' cannot be mapped to a row field name"
            )
        column = getattr(row_class, row_field_name)
        if (
            isinstance(filter, StringSetFilter)
            or isinstance(filter, NumberSetFilter)
            or isinstance(filter, UuidSetFilter)
        ):
            return (
                column.in_(filter.members)
                if not invert
                else sa.not_(column.in_(filter.members))
            )
        elif isinstance(filter, ExistsFilter):
            return column != None if not invert else column == None
        elif isinstance(filter, EqualsFilter):
            return column == filter.value if not invert else column != filter.value
        elif isinstance(filter, RangeFilter):
            args = []
            if filter.lower_bound:
                if filter.lower_bound_censor == ComparisonOperator.GT:
                    args.append(column > filter.lower_bound)
                elif filter.lower_bound_censor == ComparisonOperator.GTE:
                    args.append(column >= filter.lower_bound)
            if filter.upper_bound:
                if filter.upper_bound_censor == ComparisonOperator.ST:
                    args.append(column < filter.upper_bound)
                elif filter.upper_bound_censor == ComparisonOperator.STE:
                    args.append(column <= filter.upper_bound)
            if len(args) == 1:
                return args[0] if not invert else sa.not_(args[0])
            return sa.and_(*args) if not invert else sa.not_(sa.and_(*args))
        raise exc.InvalidArgumentsError(
            f"Unsupported filter type: {filter.__class__.__name__}"
        )

    def _split_filter_recursion(
        self, field_name_map: dict[str, str], filter: Filter
    ) -> tuple[Filter | None, Filter | None]:
        map_key_only_classes = [
            ExistsFilter,
            EqualsBooleanFilter,
            EqualsNumberFilter,
            EqualsStringFilter,
            EqualsUuidFilter,
            StringSetFilter,
            NumberSetFilter,
            UuidSetFilter,
            DateRangeFilter,
            DatetimeRangeFilter,
            NumberRangeFilter,
        ]
        # Convert composite filter if possible
        if isinstance(filter, CompositeFilter):
            where_clause_filters = []
            remainder_filters = []
            if filter.operator == LogicalOperator.OR:
                # Split only when all sub-filters can fully be converted into a where
                # clause
                for sub_filter in filter.filters:
                    where_clause_filter, remainder_filter = (
                        self._split_filter_recursion(field_name_map, sub_filter)
                    )
                    if remainder_filter is not None:
                        # Subfilter could not be converted completely -> filter cannot
                        # be converted
                        return None, filter
                    where_clause_filters.append(where_clause_filter)
                return (
                    CompositeFilter(
                        filters=where_clause_filters, operator=LogicalOperator.OR
                    ),
                    None,
                )
            if filter.operator == LogicalOperator.AND:
                # Split all sub-filters
                for sub_filter in filter.filters:
                    where_clause_filter, remainder_filter = (
                        self._split_filter_recursion(field_name_map, sub_filter)
                    )
                    if where_clause_filter:
                        where_clause_filters.append(where_clause_filter)
                    if remainder_filter:
                        remainder_filters.append(remainder_filter)
                # Combine where clause and remainder filters each into a single filter
                if len(where_clause_filters) == 0:
                    where_clause_filter = None
                elif len(where_clause_filters) == 1:
                    where_clause_filter = where_clause_filters[0]
                else:
                    where_clause_filter = CompositeFilter(
                        filters=where_clause_filters, operator=LogicalOperator.AND
                    )
                if len(remainder_filters) == 0:
                    remainder_filter = None
                elif len(remainder_filters) == 1:
                    remainder_filter = remainder_filters[0]
                else:
                    remainder_filter = CompositeFilter(
                        filters=remainder_filters, operator=LogicalOperator.AND
                    )
                return where_clause_filter, remainder_filter
            # Filter cannot be converted due to unsupported operator
            return None, filter
        # Convert non-composite filter if possible
        mapped_key = field_name_map.get(filter.get_key())
        if not mapped_key:
            # Field name cannot be mapped
            return None, filter
        for filter_class in map_key_only_classes:
            if isinstance(filter, filter_class):
                values = filter.model_dump()
                values["key"] = mapped_key
                return filter_class(**values), None
        # Filter cannot be converted
        return None, filter

    def print_db_content(self, model_class: type[Model], **kwargs: Any) -> None:
        """Helper method for debugging"""
        header = kwargs.get("header", "")
        mapper = self.get_mapper(model_class)
        tables_classes = [
            mapper.row_class,
        ]
        row_sets = []
        with self.get_session() as session:
            for table_class in tables_classes:
                row_sets.append(list(session.query(table_class)) if table_class else [])
            session.commit()
            for table_class, row_set in zip(tables_classes, row_sets):
                if not table_class:
                    continue
                if not row_set:
                    print(f"{header}empty {table_class}")
                for row in row_set:
                    print(f"{header}{row}")

    def _in_session_add_cascade_read(
        self,
        session: Session,
        links: dict[int, Link],
        objs: list[Model],
        optimize_parameter_handling: bool = False,
    ) -> None:
        # Go over each link
        for link in links.values():
            # Get unique link ids to retrieve
            link_mapper = self.get_mapper(link.link_model_class)
            link_ids = [getattr(x, link.link_field_name) for x in objs]
            uq_link_ids = set(link_ids)
            uq_link_ids.discard(None)
            # Retrieve unique link objs
            uq_link_rows, uq_link_ids = SARepository._in_session_read_some(
                link_mapper,
                session,
                link_mapper.row_class,
                list(uq_link_ids),
                optimize_parameter_handling,
            )
            uq_link_objs = self.from_sql(link.link_model_class, uq_link_rows)
            # Map link objs to ids and set in objs
            uq_link_objs = dict(zip(uq_link_ids, uq_link_objs))
            uq_link_objs[None] = None
            for obj, link_id in zip(objs, link_ids):
                setattr(
                    obj,
                    link.relationship_field_name,
                    uq_link_objs[link_id],
                )

    def verify_valid_ids(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable,
        model_class: type[Model],
        obj_ids: Iterable[Hashable],
        verify_exists: bool = True,
        verify_duplicate: bool = True,
    ) -> None:
        # Check arguments
        if not verify_exists and not verify_duplicate:
            return
        if not isinstance(obj_ids, list):
            obj_ids = list(obj_ids)
        obj_ids_set = set(obj_ids)
        if verify_duplicate and len(obj_ids) != len(obj_ids_set):
            seen = set()
            uq_obj_ids = set(
                x for x in obj_ids if x not in seen and not seen.add(x)  # type: ignore[func-returns-value]
            )
            duplicate_obj_ids = obj_ids_set - uq_obj_ids
            raise exc.DuplicateIdsError("obj_ids is not unique", ids=duplicate_obj_ids)
        # Verify existence of objs
        if verify_exists:
            try:
                self.crud(
                    uow,
                    user_id,
                    model_class,
                    None,
                    list(obj_ids_set),
                    CrudOperation.READ_SOME,
                )
            except exc.InvalidIdsError as e:
                # TODO: determine invalid obj_ids and pass them to the exception
                raise exc.InvalidIdsError("Invalid obj_ids", ids=None) from e

    @staticmethod
    def _select_with_id_join(
        session: Session,
        get_row_id: Callable[[type], sa.Column],
        row_class: type,
        obj_ids: list[Hashable],
    ) -> sa.sql.Select:
        """
        Implement a SELECT statement with an INNER JOIN to restrict to the obj_ids passed
        using dialect-specific temporary table creation. Concept is generic and can
        be implemented with essentially any SQL dialect but implementation specifics
        vary; at present, only MS SQL Server is supported.
        """

        dialect = session.get_bind().dialect
        if dialect.name == "mssql":
            # TODO: check if temp table exists and take a different name in that case
            temp_table_name = f"#temp_{str(uuid.uuid4()).replace('-','_')}"
            id_col_name = get_row_id(row_class).name
            id_datatype = row_class.__table__.c[id_col_name].type
            id_datatype_sql = id_datatype.compile(dialect=dialect)
            # TODO: finalize this part
            # Create the temp table
            # we might think to introspect after CREATE TABLE, but that opens us up to session/database sync and lock issues...
            # which we did experience in testing
            temp_table_obj = sa.Table(
                temp_table_name, row_class.metadata, sa.Column(id_col_name, id_datatype)
            )
            session.execute(
                sa.text(
                    f"CREATE TABLE {temp_table_name} ({id_col_name} {id_datatype_sql})"
                )
            )
            # session.flush()  # need to be able to introspect!
            # temp_table_obj = sa.Table(temp_table_name, row_class.metadata, autoload_with=session.get_bind().engine)
            # hard-coded batch size; MS SQL Server limit is 2,100; we just use something reasonable
            # no urgent need to turn hard-coding into a parameter as this is dialect-specific issues
            # handled in dialect-specific code
            batch_size = 1000
            for i in range(0, len(obj_ids), batch_size):
                oid_batch = obj_ids[i : i + batch_size]
                values = [{id_col_name: v} for v in oid_batch]
                session.execute(sa.insert(temp_table_obj), values)
                session.flush()
        else:
            raise NotImplementedError(
                "Only MS SQL Server is supported by _select_with_id_join"
            )

        # Select with join to restrict to ids passed
        sql_select = select(row_class).join(
            temp_table_obj,
            row_class.__table__.c[id_col_name] == temp_table_obj.c[id_col_name],
        )
        return sql_select

    @staticmethod
    def _in_session_read_some(
        mapper: BaseSAMapper,
        session: Session,
        row_class: type,
        obj_ids: list[Hashable],
        optimize_parameter_handling: bool = False,
    ) -> tuple[list[Any], list[Hashable]]:
        """
        :param optimize_parameter_handling: if True, avoid parameterized query that using SQL's IN that is
           more many parameters nonperformant by creating and joining with a temporary table instead
           default = False, but possibly recommend to set dynamically based on number of parameters
        """
        # n = len(obj_ids)
        # Get rows as list[(Row,)], convert to list[Row]
        get_row_id = mapper.get_row_id
        if obj_ids and optimize_parameter_handling:
            # TODO: finalize this part, remove the example
            # One approach to optmization relative to parameterized query with many params
            # is to create CTE and join with it; this improves performance relative to
            # parameterized query with many params and avoids dialect issues and works with
            # arbitrarily many parameters, but is less performant than temporary table
            # method.
            # Testing on IlesSampleContext:
            #   parameterized query on IlesResult: approx 20 rows per second
            #   CTE method: approx 1,400 rows per second (10k rows)
            #   temporary table method: approx 7,000 rows per second (10k rows)
            # Leaving here in case anyone wants to revisit
            # create a SQLAlchemy CTE from the obj_ids passed as parameters
            # union_stmt = sa.union_all(*[select(sa.literal(oid).label("obj_ids")) for oid in obj_ids])
            # cte = union_stmt.cte("cte_parms")
            # id_col_name = get_row_id(row_class).name
            # sql_select = select(row_class).join(cte, row_class.__table__.c[id_col_name] == cte.c.obj_ids)

            # Or.... the temporary table method
            sql_select = SARepository._select_with_id_join(
                session, get_row_id, row_class, obj_ids
            )
        else:
            sql_select = select(row_class).where(get_row_id(row_class).in_(obj_ids))

        rows = session.execute(sql_select).all()
        rows = [x[0] for x in rows]
        # Further process rows
        row_ids = [get_row_id(x) for x in rows]
        SARepository._in_session_verify_retrieved_ids(mapper, obj_ids, row_ids)
        return rows, row_ids

    def _execute_sa(self, session: Session, execute_fn: Callable, kwargs: dict) -> Any:
        if session:
            retval = execute_fn(session)
        else:
            with self.uow(**kwargs) as uow:
                retval = execute_fn(uow.session)
        return retval

    @staticmethod
    def _in_session_verify_retrieved_ids(
        mapper: BaseSAMapper,
        obj_ids: list[Hashable],
        row_ids: list[Hashable],
        table_name: str | None = None,
    ) -> None:
        n = len(obj_ids)
        if len(row_ids) < n:
            not_found_obj_ids = [x for x in obj_ids if x not in row_ids]
            not_found_obj_ids_str = ", ".join([f"{x}" for x in not_found_obj_ids])
            table_name = table_name or mapper.table_name
            if n == 1:
                raise exc.InvalidIdsError(
                    f"Table {table_name}: no row found for id {not_found_obj_ids_str}"
                )
            raise exc.InvalidIdsError(
                f"Table {table_name}: no rows found for ids {not_found_obj_ids_str}"
            )

    @staticmethod
    def _verify_duplicate_ids(model_class: type, obj_ids: Iterable[Hashable]) -> None:
        if not isinstance(obj_ids, list) and not isinstance(obj_ids, set):
            obj_ids = [obj_ids]
        seen = set()
        uq_obj_ids = set(
            x for x in obj_ids if x not in seen and not seen.add(x)  # type: ignore
        )
        if len(uq_obj_ids) == len(obj_ids):
            return
        duplicate_ids = set(obj_ids) - uq_obj_ids
        duplicate_ids_str = ", ".join([str(x) for x in duplicate_ids])
        raise exc.DuplicateIdsError(
            f"Model {model_class}: object ids are not unique: {duplicate_ids_str}",
            ids=duplicate_ids_str,
        )

    @classmethod
    def create_sa_repository(
        cls,
        entities: list[Entity],
        connection_string: str,
        **kwargs: Any,
    ) -> "SARepository":
        # Parse arguments
        echo = kwargs.pop("echo", False)
        register_mappers = kwargs.pop("register_mappers", True)
        recreate_sqlite_file = kwargs.pop("recreate_sqlite_file", False)
        schema_names = {x.schema_name for x in entities if x.persistable}

        is_sqlite = str(connection_string).lower().startswith("sqlite:///")
        if is_sqlite:
            sqlite_file = Path(
                re.sub(".*sqlite:///", "", connection_string, flags=re.IGNORECASE)
            )
            if recreate_sqlite_file:
                # Remove existing file
                if sqlite_file.is_file():
                    sqlite_file.unlink()
                # Create the file by creating a connection
                engine = sa.create_engine(
                    f"sqlite:///{sqlite_file.as_posix()}", echo=echo
                )
                conn = engine.connect()
                conn.close()
            elif not sqlite_file.is_file():
                raise ValueError("Unable to derive file from connection string")

            # Filter some warnings
            warnings.filterwarnings(
                "ignore",
                r"^Dialect sqlite\+pysqlite does not support updated rowcount.*",
                sa.exc.SAWarning,
            )

            # Create engine, creating the sqlite file(s) if needed
            engine = sa.create_engine("sqlite:///:memory:", echo=echo)

            # Make sure foreign key constraints are enforced,
            # which is not the default for sqlite
            @sa.event.listens_for(engine, "connect")
            def set_sqlite_pragma(
                dbapi_connection: Any, connection_record: Any
            ) -> None:
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")

                cursor.close()

            # Add each schema as a separate database, as sqlite does not support schemas
            with engine.connect() as conn:
                if len(schema_names) > 1:
                    raise NotImplementedError(
                        "Multiple schemas: " + ", ".join(schema_names)
                    )
                for schema_name in schema_names:
                    conn.execute(
                        sa.text(f"attach database '{sqlite_file}' as '{schema_name}';")
                    )
        else:
            engine = EngineFactory.create_engine(connection_string, echo)

            # Create schemas if not exists
            for schema_name in schema_names:
                if not schema_name:
                    continue
                with engine.connect() as conn:
                    # print(conn)
                    result = conn.execute(sa.text("SELECT name FROM sys.schemas"))
                    schemas = [row[0] for row in result]
                    # print(schemas)
                    conn.dialect
                    if not conn.dialect.has_schema(conn, schema_name):
                        conn.execute(sa.schema.CreateSchema(schema_name))
                        conn.commit()

        # Create all tables if necessary
        metadata_set = set()
        for entity in entities:
            if not entity.persistable:
                continue
            db_model_class = entity.db_model_class
            metadata_set.add(db_model_class.metadata)

        for metadata in metadata_set:
            metadata.create_all(engine)

        # Create repository
        repository = cls(
            engine, entities=entities, register_mappers=register_mappers, **kwargs
        )

        return repository

    @classmethod
    def test_connection(
        cls,
        connection_string: str,
        **kwargs: Any,
    ) -> BaseException | None:
        try:
            connection = sa.create_engine(
                connection_string,
                connect_args=kwargs,
            ).connect()
            connection.close()
            return None
        except BaseException as exception:
            # Connection failed, skip loading
            return exception



import datetime
from typing import Any
from uuid import UUID

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, exc, model
from gen_epix.commondb.domain.service.organization import BaseOrganizationService
from gen_epix.commondb.domain.service.rbac import BaseRbacService
from gen_epix.fastapp import BaseUnitOfWork, BaseUserManager, CrudOperation, Permission
from gen_epix.fastapp.services.auth import get_email_from_claims
from gen_epix.fastapp.services.auth.util import get_name_from_claims


class UserManager(BaseUserManager):
    DEFAULT_KEY_CLAIM = "__key__"
    DEFAULT_NAME_CLAIMS: list[str | list[str]] = [
        "name",
        ["first_name", "last_name"],
        "preferred_username",
        "preferredUsername",
        "username",
    ]

    def __init__(
        self,
        organization_service: BaseOrganizationService,
        rbac_service: BaseRbacService,
        root_cfg: dict[str, dict[str, str]],
        automatic_new_user_cfg: dict[str, dict[str, str]] | None = None,
        key_claim: str | None = None,
        name_claims: list[str | list[str]] | None = None,
    ):
        # Assign input properties
        self._organization_service = organization_service
        self._rbac_service = rbac_service
        self._key_claim = key_claim or self.DEFAULT_KEY_CLAIM
        self._name_claims = name_claims or self.DEFAULT_NAME_CLAIMS

        # Derive some properties
        app_impl: AppImplDetails = organization_service.app.impl
        self._user_class: type[model.User] = app_impl.get_mapped_class(model.User)
        self._user_invitation_class: type[model.UserInvitation] = (
            app_impl.get_mapped_class(model.UserInvitation)
        )

        # Generate root organization and user objects
        self._root_organization = model.Organization(
            **root_cfg["organization"]  # type:ignore[arg-type]
        )
        if self._root_organization.id is None:
            raise exc.InitializationServiceError(
                "Root organization ID is not set in the configuration"
            )
        self._root_user = self._user_class(
            is_active=True,
            organization_id=self._root_organization.id,
            roles={self._rbac_service.root_role},
            **root_cfg["user"],  # type:ignore[arg-type]
        )

        # Get automatic new user data
        self._automatic_new_user: dict[str, Any] | None = None
        if automatic_new_user_cfg:
            self._automatic_new_user = {}
            self._automatic_new_user["organization"] = dict(
                automatic_new_user_cfg["organization"]
            )
            self._automatic_new_user["roles"] = set(automatic_new_user_cfg["roles"])
            if "id" not in self._automatic_new_user["organization"]:
                raise exc.InitializationServiceError(
                    "Automatic new user organization ID is not set in the configuration"
                )
            # fmt:off
            self._automatic_new_user["organization"]["id"] = (  # pyright:ignore[reportArgumentType]
                UUID(  
                    self._automatic_new_user["organization"]["id"]
                )
            )
            # fmt:on

    def generate_id(self) -> UUID:
        return self._organization_service.generate_id()  # type:ignore[return-value]

    def get_user_key_from_claims(self, claims: dict[str, Any]) -> str | None:
        return claims.get(self._key_claim)

    def get_user_name_from_claims(self, claims: dict[str, Any]) -> str | None:
        return get_name_from_claims(claims, self._name_claims)

    def get_user_instance_from_claims(
        self, claims: dict[str, Any]
    ) -> model.User | None:
        if self._automatic_new_user is None:
            return None
        roles = (
            self._automatic_new_user["roles"]
            if self._automatic_new_user
            else {self._rbac_service.guest_role}
        )
        organization_id = (
            self._automatic_new_user["organization"]["id"]
            if self._automatic_new_user
            else self._root_organization.id
        )
        key = claims.get(self._key_claim)
        if not key:
            raise exc.CredentialsAuthError("Key not found in claims")
        email = get_email_from_claims(claims)
        name = self.get_user_name_from_claims(claims)
        return self._user_class(
            key=key,
            email=email,
            name=name,
            is_active=True,
            roles=roles,
            organization_id=organization_id,
        )

    def is_root_user_claims(self, claims: dict[str, Any]) -> bool:
        return self._root_user.key == claims.get(self._key_claim)

    def is_root_user(self, user: model.User) -> bool:  # type:ignore[override]
        return self._rbac_service.root_role in user.roles

    def create_root_user_from_claims(self, claims: dict[str, Any]) -> model.User:
        assert self._organization_service.repository

        # Handle transactions
        with self._organization_service.repository.uow() as uow:
            # Create root organization if necessary
            cfg_root_organization: model.Organization = self._root_organization
            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                cfg_root_organization.id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                _ = self._organization_service.repository.crud(
                    uow,
                    None,
                    model.Organization,
                    cfg_root_organization,
                    None,
                    CrudOperation.CREATE_ONE,
                )

            # Create root user
            is_existing_root_user = (
                self._organization_service.repository.is_existing_user_by_key(
                    uow, claims.get(self._key_claim)
                )
            )
            if is_existing_root_user:
                raise exc.InitializationServiceError(
                    "Root user with the specified key already exists"
                )
            # Create and store root user
            root_user = self._root_user.model_copy()
            root_user.id = (
                self._organization_service.generate_id()
            )  # type:ignore[assignment]
            root_user.email = get_email_from_claims(claims)
            root_user.name = self.get_user_name_from_claims(claims)
            user: model.User = (
                self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    root_user.id,
                    self._user_class,
                    root_user,
                    None,
                    CrudOperation.CREATE_ONE,
                )
            )

        return user

    def create_user_from_claims(self, claims: dict[str, Any]) -> model.User | None:
        if self._automatic_new_user is None:
            return None
        assert self._organization_service.repository
        organization_id = self._automatic_new_user["organization"]["id"]
        with self._organization_service.repository.uow() as uow:
            # Verify if organization exists
            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                raise exc.InitializationServiceError(
                    "Automatic new user organization does not exist"
                )

            # Verify if user exists and add if not
            # TODO: refactor this to add a separate method for a potential existing user
            is_existing_user = (
                self._organization_service.repository.is_existing_user_by_key(
                    uow, claims.get(self._key_claim)
                )
            )
            if is_existing_user:
                raise exc.ServiceException(
                    f"User with key {claims.get(self._key_claim)} already exists"
                )
            claims_user = self.get_user_instance_from_claims(claims)
            if not claims_user:
                raise exc.ServiceException(
                    f"Unable to create user with key {claims.get(self._key_claim)} from claims"
                )
            claims_user.id = self.generate_id()
            user: model.User = (
                self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    claims_user.id,
                    self._user_class,
                    claims_user,
                    None,
                    CrudOperation.CREATE_ONE,
                )
            )

            # Add user case policies by calling switching organization method
            try:
                user = self._organization_service.app.handle(
                    command.UpdateUserOwnOrganizationCommand(
                        user=user,
                        organization_id=user.organization_id,
                        is_new_user=True,
                    ),
                )
            except Exception as exception:
                raise exc.UnauthorizedAuthError("Unable to add user case policies")

        return user

    def create_new_user_from_token(  # type:ignore[override]
        self, user: model.User, token: str, **kwargs: Any
    ) -> model.User:
        assert self._organization_service.repository
        created_by_user_id: UUID = kwargs["created_by_user_id"]

        with self._organization_service.repository.uow() as uow:
            # Verify if create_by_user exists and is active
            is_existing_user = self._organization_service.repository.crud(
                uow,
                None,
                self._user_class,
                None,
                created_by_user_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_user:
                raise exc.UnauthorizedAuthError("Created by user does not exist")
            created_by_user = self.retrieve_user_by_id(created_by_user_id)
            if not created_by_user.is_active:
                raise exc.UnauthorizedAuthError("Created by user is not active")

            # Verify if create_by_user made an invitation for this user that is valid
            timestamp = datetime.datetime.now()
            user_invitations: list[model.UserInvitation] = (
                self._organization_service.repository.crud(  # type: ignore[assignment]
                    uow,
                    created_by_user_id,
                    self._user_invitation_class,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
            )

            # At least one invitation exists matching the criteria
            user_invitations = [
                x
                for x in user_invitations
                if x.invited_by_user_id == created_by_user_id
                and x.token == token
                and x.key == user.key
                and x.organization_id == user.organization_id
                and x.expires_at > timestamp
            ]
            if not user_invitations:
                raise exc.UnauthorizedAuthError("Invitation does not exist")

            # Verify if organization exists
            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                user.organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                raise exc.UnauthorizedAuthError("Organization does not exist")

            is_existing_user = self.is_existing_user_by_key(user.email, uow)
            if is_existing_user:
                raise exc.UnauthorizedAuthError("User already exists")

            try:
                created_user: model.User = (
                    self._organization_service.repository.crud(  # type:ignore[assignment]
                        uow,
                        created_by_user_id,
                        self._user_class,
                        self._user_class(
                            **(user.model_dump() | {"id": self.generate_id()})
                        ),
                        None,
                        CrudOperation.CREATE_ONE,
                    )
                )
            except:
                raise exc.UnauthorizedAuthError("Unable to create user")

            return created_user

    def is_existing_user_by_key(
        self, user_key: str | None, uow: BaseUnitOfWork
    ) -> bool:
        return self._organization_service.repository.is_existing_user_by_key(
            uow, user_key
        )

    def retrieve_user_by_key(self, user_key: str) -> model.User:
        return self._organization_service.retrieve_user_by_key(user_key)

    def retrieve_user_by_id(self, user_id: UUID) -> model.User:  # type:ignore[override]
        with self._organization_service.repository.uow() as uow:
            user: model.User = (
                self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id,
                    self._user_class,
                    None,
                    user_id,
                    CrudOperation.READ_ONE,
                )
            )
        return user

    def update_user_name(  # type:ignore[override]
        self, user: model.User, new_name: str
    ) -> model.User | None:
        if user.name == new_name:
            return user
        user.name = new_name
        with self._organization_service.repository.uow() as uow:
            updated_user: model.User = (
                self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    user.id,
                    self._user_class,
                    user,
                    None,
                    CrudOperation.UPDATE_ONE,
                )
            )
        return updated_user

    def retrieve_user_permissions(  # type:ignore[override]
        self, user: model.User
    ) -> set[Permission]:
        return self._rbac_service.retrieve_user_permissions(user)



from typing import Any

from gen_epix.commondb.policies import (
    ReadOrganizationResultsOnlyPolicy as CommonReadOrganizationResultsOnlyPolicy,
)
from gen_epix.omopdb.domain.service import BaseAbacService


class ReadOrganizationResultsOnlyPolicy(CommonReadOrganizationResultsOnlyPolicy):
    def __init__(
        self,
        abac_service: BaseAbacService,
        **kwargs: Any,
    ):
        super().__init__(
            abac_service,
            **kwargs,
        )
        self.has_organization_id_attr_command_classes.update(set())
        self.has_user_id_attr_command_classes.update(set())



# pylint: disable=useless-import-alias
from gen_epix.fastapp.services.auth.base import BaseAuthService as BaseAuthService
from gen_epix.fastapp.services.auth.command import (
    GetIdentityProvidersCommand as GetIdentityProvidersCommand,
)
from gen_epix.fastapp.services.auth.idp_client import IdpClient as IdpClient
from gen_epix.fastapp.services.auth.literal import EMAIL_PATTERN as EMAIL_PATTERN
from gen_epix.fastapp.services.auth.mock_idp_client import (
    MockIDPClient as MockIDPClient,
)
from gen_epix.fastapp.services.auth.model import Claims as Claims
from gen_epix.fastapp.services.auth.model import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth.model import IDPUser as IDPUser
from gen_epix.fastapp.services.auth.model import OidcServerCfg as OidcServerCfg
from gen_epix.fastapp.services.auth.oauth_idp_client import (
    OauthIdpClient as OauthIdpClient,
)
from gen_epix.fastapp.services.auth.service import AuthService as AuthService
from gen_epix.fastapp.services.auth.util import (
    get_email_from_claims as get_email_from_claims,
)



import abc
import datetime
import json
from typing import Any


class BaseLogItem(abc.ABC):
    """
    BaseLogItem class for creating log messages. Defined as a regular class instead of a
    dataclass for efficiency reasons. The `dumps` method is used to convert the object
    to a JSON string that can be inserted in a log.
    """

    def __init__(self, **kwargs: Any) -> None:
        self.content = kwargs

    @abc.abstractmethod
    def dumps(
        self, indent: int | str | None = None, separators: tuple[str, str] = (",", ":")
    ) -> str:
        raise NotImplementedError

    @staticmethod
    def _custom_json_encoder(obj: Any) -> str:
        if isinstance(obj, Exception):
            # TODO: Provide more structured encoding of exception
            return str(obj)
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return str(obj)


class LogItem(BaseLogItem):
    def __init__(self, **kwargs: Any) -> None:
        self.code: str | None = kwargs.pop("code", None)  # type: ignore
        self.msg: str | None = kwargs.pop("msg", None)  # type: ignore
        self.content = kwargs if kwargs else None

    def dumps(self, indent=None, separators=(",", ":")) -> str:
        msg = {
            "code": self.code,
            "msg": self.msg,
        }
        msg = msg if not self.content else msg | self.content
        return json.dumps(
            msg,
            indent=indent,
            separators=separators,
            default=BaseLogItem._custom_json_encoder,
        )



from gen_epix.fastapp.repositories.sa.repository import SARepository
from gen_epix.seqdb.domain.repository.file import BaseFileRepository


class FileSARepository(SARepository, BaseFileRepository):
    pass



from typing import ClassVar

import gen_epix.commondb.domain.model.abac as model
from gen_epix.commondb.domain.command.base import Command, CrudCommand

# Non-CRUD


class RetrieveOrganizationsUnderAdminCommand(Command):
    """
    Retrieve the ids of all the organizations under administration by the user
    executing the command.
    """

    pass


# CRUD


class OrganizationAdminPolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationAdminPolicy



import abc
from enum import Enum
from typing import Any

from gen_epix.commondb.domain import command, enum
from gen_epix.fastapp.app import App
from gen_epix.fastapp.model import Permission
from gen_epix.fastapp.services.rbac import BaseRbacService as ServiceBaseRbacService


class BaseRbacService(ServiceBaseRbacService):
    SERVICE_TYPE = enum.ServiceType.RBAC

    def __init__(self, app: App, **kwargs: Any) -> None:
        super().__init__(app, **kwargs)
        self.role_map: dict[enum.Role | Enum, str]
        self.role_set_map: dict[enum.RoleSet | Enum, frozenset[str]]
        self.root_role: str
        self.guest_role: str

    def register_handlers(self) -> None:
        self.register_default_crud_handlers()
        f = self.app.register_handler
        f(
            command.RetrieveOwnPermissionsCommand,
            self.retrieve_own_permissions,
        )
        f(
            command.RetrieveSubRolesCommand,
            self.retrieve_sub_roles,
        )

    @abc.abstractmethod
    def retrieve_own_permissions(
        self, cmd: command.RetrieveOwnPermissionsCommand
    ) -> set[Permission]:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_sub_roles(self, cmd: command.RetrieveSubRolesCommand) -> set[str]:
        raise NotImplementedError



from gen_epix.casedb.domain.repository import BaseSubjectRepository
from gen_epix.fastapp.repositories import DictRepository


class SubjectDictRepository(DictRepository, BaseSubjectRepository):
    pass



from typing import Any

from gen_epix.casedb.domain import DOMAIN, command, model
from gen_epix.casedb.domain.policy import RoleGenerator
from gen_epix.casedb.policies import COMMON_POLICY_MAP
from gen_epix.casedb.services import RbacService
from gen_epix.commondb.config import AppCfg
from gen_epix.commondb.env import AppComposer as CommonAppComposer


class AppComposer(CommonAppComposer):
    def __init__(
        self,
        app_cfg: AppCfg,
        log_setup: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            app_cfg,
            log_setup=log_setup,
            domain=DOMAIN,
            sorted_service_types=model.SORTED_SERVICE_TYPES,  # type: ignore[arg-type]
            model_class_map=model.COMMON_MODEL_MAP,  # type: ignore[arg-type]
            command_class_map=command.COMMON_COMMAND_MAP,  # type: ignore[arg-type]
            policy_class_map=COMMON_POLICY_MAP,
            role_generator_class=RoleGenerator,
            rbac_service_class=RbacService,
            **kwargs,
        )



"""
Multi-field transformer implementation.
"""

from collections.abc import Callable, Hashable
from typing import Any

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class MultiFieldTransformer(Transformer):
    """Transform multiple fields simultaneously."""

    def __init__(
        self,
        field_mapping: dict[Hashable, Callable[[Any], Any]],
        name: str | None = None,
    ):
        super().__init__(name)
        self.field_mapping = field_mapping

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Transform all specified fields."""
        for field_key, transform_fn in self.field_mapping.items():
            if obj.has_key(field_key):
                old_value = obj.get(field_key)
                new_value = transform_fn(old_value)
                obj.set(field_key, new_value)
        return obj



# pylint: disable=useless-import-alias
from gen_epix.casedb.repositories import sa_model as sa_model  # Initialize SA Models
from gen_epix.casedb.repositories.abac_dict import (
    AbacDictRepository as AbacDictRepository,
)
from gen_epix.casedb.repositories.abac_sa import AbacSARepository as AbacSARepository
from gen_epix.casedb.repositories.case_dict import (
    CaseDictRepository as CaseDictRepository,
)
from gen_epix.casedb.repositories.case_sa import CaseSARepository as CaseSARepository
from gen_epix.casedb.repositories.geo_dict import GeoDictRepository as GeoDictRepository
from gen_epix.casedb.repositories.geo_sa import GeoSARepository as GeoSARepository
from gen_epix.casedb.repositories.ontology_dict import (
    OntologyDictRepository as OntologyDictRepository,
)
from gen_epix.casedb.repositories.ontology_sa import (
    OntologySARepository as OntologySARepository,
)
from gen_epix.casedb.repositories.organization_dict import (
    OrganizationDictRepository as OrganizationDictRepository,
)
from gen_epix.casedb.repositories.organization_sa import (
    OrganizationSARepository as OrganizationSARepository,
)
from gen_epix.casedb.repositories.subject_dict import (
    SubjectDictRepository as SubjectDictRepository,
)
from gen_epix.casedb.repositories.subject_sa import (
    SubjectSARepository as SubjectSARepository,
)
from gen_epix.commondb.repositories import SystemDictRepository as SystemDictRepository
from gen_epix.commondb.repositories import SystemSARepository as SystemSARepository
from gen_epix.commondb.repositories.organization_dict import (
    OrganizationDictRepository as OrganizationDictRepository,
)
from gen_epix.commondb.repositories.organization_sa import (
    OrganizationSARepository as OrganizationSARepository,
)



from gen_epix.commondb.services import RbacService as CommonRbacService
from gen_epix.omopdb.domain import enum


class RbacService(CommonRbacService):
    def __init__(self, app, logger=None, **kwargs):
        super().__init__(app, logger=logger, role_enum=enum.Role, **kwargs)



import abc
from collections.abc import Iterable
from uuid import UUID

import numpy as np

from gen_epix.fastapp import BaseRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork


class BaseSeqRepository(BaseRepository):

    @abc.abstractmethod
    def get_distance_matrix_by_seq_ids(
        self,
        uow: BaseUnitOfWork,
        seq_distance_protocol_id: UUID,
        seq_ids: list[UUID],
    ) -> np.ndarray:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_seq_fasta(
        self,
        uow: BaseUnitOfWork,
        seq_ids: list[UUID],
    ) -> Iterable[tuple[str, str]]:
        raise NotImplementedError()



from typing import Any

from sqlalchemy import Engine

from gen_epix.commondb.repositories import (
    OrganizationSARepository as CommonOrganizationSARepository,
)
from gen_epix.commondb.repositories import sa_model
from gen_epix.omopdb.domain import model


class OrganizationSARepository(CommonOrganizationSARepository):
    def __init__(
        self,
        engine: Engine,
        **kwargs: Any,
    ):
        super().__init__(
            engine,
            user_class=model.User,
            user_invitation_class=model.UserInvitation,
            sa_user_class=sa_model.User,
            sa_user_invitation_class=sa_model.UserInvitation,
            **kwargs,
        )



from typing import ClassVar

from gen_epix.commondb.domain.command import CrudCommand
from gen_epix.seqdb.domain import model


class FileCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.File



# pylint: disable=useless-import-alias
from gen_epix.commondb.domain.service.abac import BaseAbacService as BaseAbacService
from gen_epix.commondb.domain.service.organization import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.commondb.domain.service.rbac import BaseRbacService as BaseRbacService
from gen_epix.commondb.domain.service.system import (
    BaseSystemService as BaseSystemService,
)
from gen_epix.commondb.domain.service.user_manager import (
    BaseUserManager as BaseUserManager,
)
from gen_epix.fastapp.services.auth import BaseAuthService as BaseAuthService



from __future__ import annotations

from uuid import UUID

from cachetools import TTLCache, cached

from gen_epix.casedb import policies as policies
from gen_epix.casedb.domain import command, exc, model
from gen_epix.casedb.domain.service.abac import BaseAbacService
from gen_epix.commondb.domain.enum import RoleSet as CommonRoleSet
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp import CrudOperation, EventTiming
from gen_epix.fastapp.model import Command
from gen_epix.filter import (
    CompositeFilter,
    EqualsBooleanFilter,
    EqualsUuidFilter,
    LogicalOperator,
    UuidSetFilter,
)


class AbacService(BaseAbacService):

    CACHE_INVALIDATION_COMMANDS: tuple[type[Command], ...] = (
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
        command.OrganizationAccessCasePolicyCrudCommand,
        command.OrganizationShareCasePolicyCrudCommand,
    )

    def register_policies(
        self,
        organization_admin_write_commands: set[type[Command]] | None = None,
        read_user_commands: set[type[Command]] | None = None,
        update_user_commands: set[type[Command]] | None = None,
        read_organization_results_only_commands: set[type[Command]] | None = None,
        read_self_results_only_commands: set[type[Command]] | None = None,
    ) -> None:
        super().register_policies(
            organization_admin_write_commands=organization_admin_write_commands,
            read_user_commands=read_user_commands,
            update_user_commands=update_user_commands,
            read_organization_results_only_commands=read_organization_results_only_commands,
            read_self_results_only_commands=read_self_results_only_commands,
        )
        policy = policies.CaseAbacPolicy(self)
        f = self.app.register_policy
        for command_class in BaseAbacService.CASE_ABAC_COMMANDS:
            f(command_class, policy, EventTiming.DURING)

    def get_case_abac(self, cmd: command.Command) -> model.CaseAbac:
        if cmd.user is None or cmd.user.id is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        user_id = cmd.user.id
        return self._get_case_abac_cached(user_id)

    def temp_update_user_own_organization(
        self,
        cmd: command.UpdateUserOwnOrganizationCommand,
    ) -> model.User:
        """
        Behaviour:
        - Update User.organization
        - Create UserAccessCasePolicies for the user and their new organization, that
          are the union of all UserAccessCasePolicies for that organization. Analogous
          for UserShareCasePolicies.

        - Delete any OrganizationAdminPolicy for the user and their previous
          organization
        - Delete any UserAccessCasePolicy for the user and their previous organization.
          Analogous for UserShareCasePolicy.
        - Create UserAccessCasePolicies for the user and their new organization, one
          for each OrganizationAccessCasePolicy for the new organization combined with
          each case type set and rights from the user's UserAccessCasePolicies from
          the previous organization. Analogous for OrganizationShareCasePolicy and
          UserShareCasePolicy.
        """
        is_new_user = cmd.is_new_user
        tgt_organization_id = cmd.organization_id
        assert cmd.user is not None
        user: model.User = cmd.user  # type: ignore[assignment]
        assert user.id is not None

        # Special case: new organization is same as current
        if user.organization_id == tgt_organization_id and not is_new_user:
            return user

        with self.repository.uow() as uow:
            # Get all current user access and share case policies
            user_access_case_policies: list[model.UserAccessCasePolicy] = (
                self.app.handle(
                    command.UserAccessCasePolicyCrudCommand(
                        user=user,
                        obj_ids=None,
                        operation=CrudOperation.READ_ALL,
                        query_filter=EqualsUuidFilter(
                            key="user_id",
                            value=user.id,
                        ),
                    )
                )
            )
            user_share_case_policies: list[model.UserShareCasePolicy] = self.app.handle(
                command.UserShareCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="user_id",
                        value=user.id,
                    ),
                )
            )
            # Get all target organization access and share case policies
            organization_access_case_policies: list[
                model.OrganizationAccessCasePolicy
            ] = self.app.handle(
                command.OrganizationAccessCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="organization_id",
                        value=tgt_organization_id,
                    ),
                )
            )
            organization_share_case_policies: list[
                model.OrganizationShareCasePolicy
            ] = self.app.handle(
                command.OrganizationShareCasePolicyCrudCommand(
                    user=user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    query_filter=EqualsUuidFilter(
                        key="organization_id",
                        value=tgt_organization_id,
                    ),
                )
            )
            # Convert target organization policies to new user policies
            new_user_access_case_policies = [
                model.UserAccessCasePolicy(
                    user_id=user.id, **x.model_dump(exclude={"id"})
                )
                for x in organization_access_case_policies
            ]
            new_user_share_case_policies = [
                model.UserShareCasePolicy(
                    user_id=user.id, **x.model_dump(exclude={"id"})
                )
                for x in organization_share_case_policies
            ]
            # Delete the current user policies
            if user_access_case_policies:
                _ = self.app.handle(
                    command.UserAccessCasePolicyCrudCommand(
                        user=user,
                        obj_ids=[
                            x.id for x in user_access_case_policies  # type:ignore[misc]
                        ],
                        operation=CrudOperation.DELETE_SOME,
                    )
                )
            if user_share_case_policies:
                _ = self.app.handle(
                    command.UserShareCasePolicyCrudCommand(
                        user=user,
                        obj_ids=[
                            x.id for x in user_share_case_policies  # type:ignore[misc]
                        ],
                        operation=CrudOperation.DELETE_SOME,
                    )
                )
            # Store the new user policies
            new_user_access_case_policies = self.app.handle(
                command.UserAccessCasePolicyCrudCommand(
                    user=user,
                    objs=new_user_access_case_policies,  # type:ignore[arg-type]
                    operation=CrudOperation.CREATE_SOME,
                )
            )
            new_user_share_case_policies = self.app.handle(
                command.UserShareCasePolicyCrudCommand(
                    user=user,
                    objs=new_user_share_case_policies,  # type:ignore[arg-type]
                    operation=CrudOperation.CREATE_SOME,
                )
            )
            # Change the user organization
            updated_user = user.model_copy()
            updated_user.organization_id = tgt_organization_id
            user = self.app.handle(
                command.UserCrudCommand(
                    user=user,
                    objs=updated_user,
                    operation=CrudOperation.UPDATE_ONE,
                )
            )

        # Invalidate cache
        # TODO: develop general system for caching and cache invalidation
        self._get_user_by_id_cached.cache_clear()  # type: ignore[attr-defined]
        self._get_case_abac_cached.cache_clear()  # type: ignore[attr-defined]

        return user

    @cached(cache=TTLCache(maxsize=1024, ttl=300))
    def _get_case_abac_cached(
        self,
        user_id: UUID,
    ) -> model.CaseAbac:
        user = self._get_user_by_id_cached(user_id)
        organization_id = user.organization_id
        # @ABAC: Special case: user has full access, defined as all active private data collection policies and all active organization access and share case policies
        is_full_access = not self.role_set_map[CommonRoleSet.GE_APP_ADMIN].isdisjoint(
            user.roles
        )
        if is_full_access:
            return model.CaseAbac(
                is_full_access=True,
                case_type_access_abacs={},
                case_type_share_abacs={},
            )
        # Get filters for all the policies
        organization_filter = CompositeFilter(
            filters=[
                EqualsBooleanFilter(key="is_active", value=True),
                EqualsUuidFilter(key="organization_id", value=organization_id),
            ],
            operator=LogicalOperator.AND,
        )
        user_filter = CompositeFilter(
            filters=[
                EqualsBooleanFilter(key="is_active", value=True),
                EqualsUuidFilter(key="user_id", value=user_id),
            ],
            operator=LogicalOperator.AND,
        )

        # Retrieve all the policies as well as the case type set members and case type col set members
        with self.repository.uow() as uow:
            # Retrieve organization access and share case policies
            organization_access_case_policies: list[
                model.OrganizationAccessCasePolicy
            ] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id=user_id,
                model_class=model.OrganizationAccessCasePolicy,
                objs=None,
                obj_ids=None,
                operation=CrudOperation.READ_ALL,
                filter=organization_filter,
            )
            organization_share_case_policies: list[
                model.OrganizationShareCasePolicy
            ] = self.repository.crud(  # type:ignore[assignment]
                uow,
                user_id=user_id,
                model_class=model.OrganizationShareCasePolicy,
                objs=None,
                obj_ids=None,
                operation=CrudOperation.READ_ALL,
                filter=organization_filter,
            )
            # Retrieve user access and share case policies
            user_access_case_policies: list[model.UserAccessCasePolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=user_id,
                    model_class=model.UserAccessCasePolicy,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=user_filter,
                )
            )
            user_share_case_policies: list[model.UserShareCasePolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=user_id,
                    model_class=model.UserShareCasePolicy,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=user_filter,
                )
            )
            # Retrieve case type cols per case type
            case_type_col_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_id, x.id)
                    for x in self.app.handle(
                        command.CaseTypeColCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                        ),
                    )
                ],
                as_set=True,
            )
            # Retrieve relevant case type set members and case type col set members
            all_case_policies: list[
                model.OrganizationAccessCasePolicy
                | model.UserAccessCasePolicy
                | model.OrganizationShareCasePolicy
                | model.UserShareCasePolicy
            ] = (
                organization_access_case_policies
                + organization_share_case_policies
                + user_access_case_policies
                + user_share_case_policies
            )
            case_type_set_ids = frozenset(x.case_type_set_id for x in all_case_policies)
            case_type_set_member_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_set_id, x.case_type_id)
                    for x in self.app.handle(
                        command.CaseTypeSetMemberCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                            query_filter=UuidSetFilter(
                                key="case_type_set_id",
                                members=case_type_set_ids,
                            ),
                        ),
                    )
                ],
                as_set=True,
            )
            all_access_case_policies: list[
                model.OrganizationAccessCasePolicy | model.UserAccessCasePolicy
            ] = (organization_access_case_policies + user_access_case_policies)
            case_type_col_set_ids: set[UUID] = set(
                x.read_case_type_col_set_id
                for x in all_access_case_policies
                if x.read_case_type_col_set_id
            ) | set(
                x.write_case_type_col_set_id
                for x in all_access_case_policies
                if x.write_case_type_col_set_id
            )
            case_type_col_set_member_map: dict[UUID, set[UUID]] = map_paired_elements(  # type: ignore[assignment]
                [
                    (x.case_type_col_set_id, x.case_type_col_id)
                    for x in self.app.handle(
                        command.CaseTypeColSetMemberCrudCommand(
                            user=user,
                            objs=None,
                            obj_ids=None,
                            operation=CrudOperation.READ_ALL,
                            query_filter=UuidSetFilter(
                                key="case_type_col_set_id",
                                members=frozenset(case_type_col_set_ids),
                            ),
                        ),
                    )
                ],
                as_set=True,
            )

        # Create and return the case abac object with the intersection of the rights
        return model.CaseAbac(
            is_full_access=is_full_access,
            case_type_access_abacs=AbacService._get_access_intersect(
                organization_access_case_policies,
                user_access_case_policies,
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            ),
            case_type_share_abacs=AbacService._get_share_intersect(
                organization_share_case_policies,
                user_share_case_policies,
                case_type_set_member_map,
            ),
        )

    @staticmethod
    def _get_access_dict(
        case_policies: list[
            model.OrganizationAccessCasePolicy | model.UserAccessCasePolicy
        ],
        case_type_set_member_map: dict[UUID, set[UUID]],
        case_type_col_map: dict[UUID, set[UUID]],
        case_type_col_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeAccessAbac]]:
        def _get_case_type_col_ids(
            case_type_col_set_id: UUID | None,
        ) -> set[UUID]:
            return (
                case_type_col_set_member_map.get(case_type_col_set_id, set())
                if case_type_col_set_id
                else set()
            )

        dict_: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = {}
        for x in case_policies:
            case_type_ids = case_type_set_member_map.get(x.case_type_set_id, set())
            is_private = (
                getattr(x, "is_private", True) if hasattr(x, "is_private") else True
            )
            # Create case type access abac object for each case type id
            for case_type_id in case_type_ids:
                if case_type_id not in dict_:
                    dict_[case_type_id] = {}
                all_case_type_col_ids = case_type_col_map.get(case_type_id, set())
                case_type_access_abac = model.CaseTypeAccessAbac(
                    case_type_id=case_type_id,
                    data_collection_id=x.data_collection_id,
                    is_private=is_private,
                    add_case=x.add_case,
                    remove_case=x.remove_case,
                    read_case_type_col_ids=all_case_type_col_ids
                    & _get_case_type_col_ids(x.read_case_type_col_set_id),
                    write_case_type_col_ids=all_case_type_col_ids
                    & _get_case_type_col_ids(x.write_case_type_col_set_id),
                    add_case_set=x.add_case_set,
                    remove_case_set=x.remove_case_set,
                    read_case_set=x.read_case_set,
                    write_case_set=x.write_case_set,
                )
                dict_[case_type_id][x.data_collection_id] = case_type_access_abac
        return dict_

    @staticmethod
    def _get_access_intersect(
        organization_access_case_policies: list[model.OrganizationAccessCasePolicy],
        user_access_case_policies: list[model.UserAccessCasePolicy],
        case_type_set_member_map: dict[UUID, set[UUID]],
        case_type_col_map: dict[UUID, set[UUID]],
        case_type_col_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeAccessAbac]]:
        dict1: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = (
            AbacService._get_access_dict(
                organization_access_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            )
        )
        dict2: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = (
            AbacService._get_access_dict(
                user_access_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
                case_type_col_map,
                case_type_col_set_member_map,
            )
        )
        dict3: dict[UUID, dict[UUID, model.CaseTypeAccessAbac]] = {}

        for case_type_id in set(dict1.keys()) & set(dict2.keys()):
            dict3[case_type_id] = {}
            data1 = dict1[case_type_id]
            data2 = dict2[case_type_id]
            for data_collection_id in set(data1.keys()) & set(data2.keys()):
                # Get both access case policies
                x = data1[data_collection_id]
                y = data2[data_collection_id]
                # Create the case type access abac object with the intersection of
                # the rights
                case_type_access_abac = model.CaseTypeAccessAbac(
                    case_type_id=case_type_id,
                    data_collection_id=data_collection_id,
                    is_private=x.is_private and y.is_private,
                    add_case=x.add_case and y.add_case,
                    remove_case=x.remove_case and y.remove_case,
                    read_case_type_col_ids=x.read_case_type_col_ids
                    & y.read_case_type_col_ids,
                    write_case_type_col_ids=x.write_case_type_col_ids
                    & y.write_case_type_col_ids,
                    add_case_set=x.add_case_set and y.add_case_set,
                    remove_case_set=x.remove_case_set and y.remove_case_set,
                    read_case_set=x.read_case_set and y.read_case_set,
                    write_case_set=x.write_case_set and y.write_case_set,
                )
                # Add to dict only if any rights remain
                if not case_type_access_abac.has_any_rights():
                    continue
                dict3[case_type_id][data_collection_id] = case_type_access_abac
        return dict3

    @staticmethod
    def _get_share_dict(
        case_policies: list[
            model.OrganizationShareCasePolicy | model.UserShareCasePolicy
        ],
        case_type_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeShareAbac]]:
        dict_: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = {}
        for x in case_policies:
            # Create case type share abac object for each case type id
            data_collection_id = x.data_collection_id
            from_data_collection_id = x.from_data_collection_id
            case_type_ids = case_type_set_member_map.get(x.case_type_set_id, set())
            for case_type_id in case_type_ids:
                if case_type_id not in dict_:
                    dict_[case_type_id] = {}
                if data_collection_id not in dict_[case_type_id]:
                    dict_[case_type_id][data_collection_id] = model.CaseTypeShareAbac(
                        case_type_id=case_type_id,
                        data_collection_id=data_collection_id,
                        add_case_from_data_collection_ids=set(),
                        remove_case_from_data_collection_ids=set(),
                        add_case_set_from_data_collection_ids=set(),
                        remove_case_set_from_data_collection_ids=set(),
                    )
                case_type_share_abac = dict_[case_type_id][data_collection_id]
                # Add the source data collection id to the appropriate sets
                if x.add_case:
                    case_type_share_abac.add_case_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.remove_case:
                    case_type_share_abac.remove_case_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.add_case_set:
                    case_type_share_abac.add_case_set_from_data_collection_ids.add(
                        from_data_collection_id
                    )
                if x.remove_case_set:
                    case_type_share_abac.remove_case_set_from_data_collection_ids.add(
                        from_data_collection_id
                    )
        return dict_

    @staticmethod
    def _get_share_intersect(
        organization_share_case_policies: list[model.OrganizationShareCasePolicy],
        user_share_case_policies: list[model.UserShareCasePolicy],
        case_type_set_member_map: dict[UUID, set[UUID]],
    ) -> dict[UUID, dict[UUID, model.CaseTypeShareAbac]]:
        dict1: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = (
            AbacService._get_share_dict(
                organization_share_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
            )
        )
        dict2: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = (
            AbacService._get_share_dict(
                user_share_case_policies,  # type:ignore[arg-type]
                case_type_set_member_map,
            )
        )
        dict3: dict[UUID, dict[UUID, model.CaseTypeShareAbac]] = {}

        for case_type_id in set(dict1.keys()) & set(dict2.keys()):
            dict3[case_type_id] = {}
            data1 = dict1[case_type_id]
            data2 = dict2[case_type_id]
            for data_collection_id in set(data1.keys()) & set(data2.keys()):
                # Get both share case policies
                x = dict1[case_type_id][data_collection_id]
                y = dict2[case_type_id][data_collection_id]
                # Create the case type share abac object with the intersection of the
                # rights
                dict3[case_type_id][data_collection_id] = model.CaseTypeShareAbac(
                    case_type_id=case_type_id,
                    data_collection_id=data_collection_id,
                    add_case_from_data_collection_ids=x.add_case_from_data_collection_ids
                    & y.add_case_from_data_collection_ids,
                    remove_case_from_data_collection_ids=x.remove_case_from_data_collection_ids
                    & y.remove_case_from_data_collection_ids,
                    add_case_set_from_data_collection_ids=x.add_case_set_from_data_collection_ids
                    & y.add_case_set_from_data_collection_ids,
                    remove_case_set_from_data_collection_ids=x.remove_case_set_from_data_collection_ids
                    & y.remove_case_set_from_data_collection_ids,
                )
        # Remove any case type share abac objects with no rights
        for case_type_id in dict3:
            to_pop_data_collection_ids = {
                x for x, y in dict3[case_type_id].items() if not y.has_any_rights()
            }
            for data_collection_id in to_pop_data_collection_ids:
                dict3[case_type_id].pop(data_collection_id)
        return dict3



# pylint: disable=wildcard-import, unused-import
# because this is a package, and imported as such in other modules

from enum import Enum

from gen_epix.commondb.domain.enum import RoleSet as RoleSet


class ServiceType(Enum):
    AUTH = "AUTH"
    ORGANIZATION = "ORGANIZATION"
    SYSTEM = "SYSTEM"
    RBAC = "RBAC"
    ABAC = "ABAC"
    GEO = "GEO"
    ONTOLOGY = "ONTOLOGY"
    SEQDB = "SEQDB"
    SUBJECT = "SUBJECT"
    CASE = "CASE"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"


class RegionRelationType(Enum):
    IS_SEPARATE_FROM = "IS_SEPARATE_FROM"
    IS_ADJACENT_TO = "IS_ADJACENT_TO"
    OVERLAPS_WITH = "OVERLAPS_WITH"
    CONTAINS = "CONTAINS"


class TreeAlgorithmType(Enum):
    # See https://en.wikipedia.org/wiki/Hierarchical_clustering
    SLINK = "SLINK"  # Single linkage clustering
    CLINK = "CLINK"  # Complete linkage clustering
    UPGMA = "UPGMA"  # Unweighted average linkage clustering
    WPGMA = "WPGMA"  # Weighted average linkage clustering
    UPGMC = "UPGMC"  # Centroid linkage clustering
    WPGMC = "WPGMC"  # Median linkage clustering
    VERSATILE = "VERSATILE"  # Versatile linkage clustering
    MISSQ = "MISSQ"  # Ward linkage, Minimum Increase of Sum of Squares
    MNSSQ = "MNSSQ"  # Minimum Error Sum of Squares
    MIVAR = "MIVAR"  # Minimum Increase in Variance
    MNVAR = "MNVAR"  # Minimum Variance
    MINI_MAX = "MINI_MAX"  # Mini-Max linkage
    HAUSDORFF = "HAUSDORFF"  # Hausdorff linkage
    MIN_SUM_MEDOID = "MIN_SUM_MEDOID"  # Minimum Sum Medoid linkage
    MIN_SUM_INCREASE_MEDOID = (
        "MIN_SUM_INCREASE_MEDOID"  # Minimum Sum Increase Medoid linkage
    )
    MEDOID = "MEDOID"  # Medoid linkage
    MIN_ENERGY = "MIN_ENERGY"  # Minimum energy clustering
    FITCH_MARGOLIASH = "FITCH_MARGOLIASH"  # Fitch–Margoliash
    MAX_PARSIMONY = "MAX_PARSIMONY"  # Maximum parsimony
    ML = "ML"  # Maximum likelihood
    BAYESIAN_INFERENCE = "BAYESIAN_INFERENCE"  # Bayesian inference
    MIN_SPANNING = "MIN_SPANNING"  # Minimum spanning
    NJ = "NJ"  # Neighbor joining


class ColRelation(Enum):
    IS_UNRELATED_TO = 0
    AGGREGATES_VALUE = 1
    AGGREGATES_COLUMN = 2


class Role(Enum):
    ROOT = "CASEDB_ROOT"
    APP_ADMIN = "CASEDB_APP_ADMIN"
    ORG_ADMIN = "CASEDB_ORG_ADMIN"
    REFDATA_ADMIN = "CASEDB_REFDATA_ADMIN"
    ORG_USER = "CASEDB_ORG_USER"
    GUEST = "CASEDB_GUEST"
    ROLE1 = "CASEDB_ROLE1"


class CaseRight(Enum):
    ADD_CASE = "ADD_CASE"
    REMOVE_CASE = "REMOVE_CASE"
    READ_CASE = "READ_CASE"
    WRITE_CASE = "WRITE_CASE"
    ADD_CASE_SET = "ADD_CASE_SET"
    REMOVE_CASE_SET = "REMOVE_CASE_SET"
    READ_CASE_SET = "READ_CASE_SET"
    WRITE_CASE_SET = "WRITE_CASE_SET"


class CaseRightSet(Enum):
    SHARE = frozenset(
        {
            CaseRight.ADD_CASE,
            CaseRight.REMOVE_CASE,
            CaseRight.ADD_CASE_SET,
            CaseRight.REMOVE_CASE_SET,
        }
    )
    CONTENT = frozenset(
        {
            CaseRight.READ_CASE,
            CaseRight.WRITE_CASE,
            CaseRight.READ_CASE_SET,
            CaseRight.WRITE_CASE_SET,
        }
    )
    ADD = frozenset(
        {
            CaseRight.ADD_CASE,
            CaseRight.ADD_CASE_SET,
        }
    )
    REMOVE = frozenset(
        {
            CaseRight.REMOVE_CASE,
            CaseRight.REMOVE_CASE_SET,
        }
    )
    CASE = frozenset(
        {
            CaseRight.ADD_CASE,
            CaseRight.REMOVE_CASE,
            CaseRight.READ_CASE,
            CaseRight.WRITE_CASE,
        }
    )
    CASE_CONTENT = frozenset(
        {
            CaseRight.READ_CASE,
            CaseRight.WRITE_CASE,
        }
    )
    CASE_SET = frozenset(
        {
            CaseRight.ADD_CASE_SET,
            CaseRight.REMOVE_CASE_SET,
            CaseRight.READ_CASE_SET,
            CaseRight.WRITE_CASE_SET,
        }
    )
    CASE_SET_CONTENT = frozenset(
        {
            CaseRight.READ_CASE_SET,
            CaseRight.WRITE_CASE_SET,
        }
    )


class CaseClassification(Enum):
    POSSIBLE = "POSSIBLE"
    PROBABLE = "PROBABLE"
    CONFIRMED = "CONFIRMED"


class CaseTypeSetCategoryPurpose(Enum):
    CONTENT = "CONTENT"
    SECURITY = "SECURITY"


class ConceptSetType(Enum):
    CONTEXT_FREE_GRAMMAR_JSON = "CONTEXT_FREE_GRAMMAR_JSON"
    CONTEXT_FREE_GRAMMAR_XML = "CONTEXT_FREE_GRAMMAR_XML"
    REGULAR_LANGUAGE = "REGULAR_LANGUAGE"
    NOMINAL = "NOMINAL"
    ORDINAL = "ORDINAL"
    INTERVAL = "INTERVAL"


class ConceptRelationType(Enum):
    CONTAINS = "CONTAINS"


class DimType(Enum):
    TEXT = "TEXT"
    IDENTIFIER = "IDENTIFIER"
    NUMBER = "NUMBER"
    TIME = "TIME"
    GEO = "GEO"
    ORGANIZATION = "ORGANIZATION"
    OTHER = "OTHER"


class ColType(Enum):
    TEXT = "TEXT"
    CONTEXT_FREE_GRAMMAR_JSON = "CONTEXT_FREE_GRAMMAR_JSON"
    CONTEXT_FREE_GRAMMAR_XML = "CONTEXT_FREE_GRAMMAR_XML"
    REGULAR_LANGUAGE = "REGULAR_LANGUAGE"
    NOMINAL = "NOMINAL"
    ORDINAL = "ORDINAL"
    INTERVAL = "INTERVAL"
    TIME_DAY = "TIME_DAY"
    TIME_WEEK = "TIME_WEEK"
    TIME_MONTH = "TIME_MONTH"
    TIME_QUARTER = "TIME_QUARTER"
    TIME_YEAR = "TIME_YEAR"
    GEO_LATLON = "GEO_LATLON"
    GEO_REGION = "GEO_REGION"
    ID_DIRECT = "ID_DIRECT"
    ID_PSEUDONYMISED = "ID_PSEUDONYMISED"
    ID_ANONYMISED = "ID_ANONYMISED"
    DECIMAL_0 = "DECIMAL_0"
    DECIMAL_1 = "DECIMAL_1"
    DECIMAL_2 = "DECIMAL_2"
    DECIMAL_3 = "DECIMAL_3"
    DECIMAL_4 = "DECIMAL_4"
    DECIMAL_5 = "DECIMAL_5"
    DECIMAL_6 = "DECIMAL_6"
    GENETIC_READS = "GENETIC_READS"
    GENETIC_SEQUENCE = "GENETIC_SEQUENCE"
    GENETIC_PROFILE = "GENETIC_PROFILE"
    GENETIC_DISTANCE = "GENETIC_DISTANCE"
    ORGANIZATION = "ORGANIZATION"
    OTHER = "OTHER"


class ColTypeSet(Enum):
    ID = frozenset({ColType.ID_DIRECT, ColType.ID_PSEUDONYMISED, ColType.ID_ANONYMISED})
    LANGUAGE = frozenset(
        {
            ColType.CONTEXT_FREE_GRAMMAR_JSON,
            ColType.CONTEXT_FREE_GRAMMAR_XML,
            ColType.REGULAR_LANGUAGE,
        }
    )
    CONTEXT_FREE_GRAMMAR = frozenset(
        {
            ColType.CONTEXT_FREE_GRAMMAR_JSON,
            ColType.CONTEXT_FREE_GRAMMAR_XML,
        }
    )
    ENTITY = frozenset(
        {
            ColType.GENETIC_SEQUENCE,
            ColType.GENETIC_DISTANCE,
            ColType.ORGANIZATION,
        }
    )
    REGULAR_LANGUAGE = frozenset(
        {
            ColType.REGULAR_LANGUAGE,
            ColType.NOMINAL,
            ColType.ORDINAL,
            ColType.INTERVAL,
            ColType.TIME_DAY,
            ColType.TIME_WEEK,
            ColType.TIME_MONTH,
            ColType.TIME_QUARTER,
            ColType.TIME_YEAR,
            ColType.GEO_LATLON,
            ColType.GEO_REGION,
            ColType.DECIMAL_0,
            ColType.DECIMAL_1,
            ColType.DECIMAL_2,
            ColType.DECIMAL_3,
            ColType.DECIMAL_4,
            ColType.DECIMAL_5,
            ColType.DECIMAL_6,
        }
    )
    STRING_SET = frozenset(
        {
            ColType.NOMINAL,
            ColType.ORDINAL,
            ColType.INTERVAL,
        }
    )
    TIME = frozenset(
        {
            ColType.TIME_DAY,
            ColType.TIME_WEEK,
            ColType.TIME_MONTH,
            ColType.TIME_QUARTER,
            ColType.TIME_YEAR,
        }
    )
    GEO = frozenset({ColType.GEO_LATLON, ColType.GEO_REGION})
    NUMBER = frozenset(
        {
            ColType.DECIMAL_0,
            ColType.DECIMAL_1,
            ColType.DECIMAL_2,
            ColType.DECIMAL_3,
            ColType.DECIMAL_4,
            ColType.DECIMAL_5,
            ColType.DECIMAL_6,
        }
    )
    GENETIC = frozenset(
        {
            ColType.GENETIC_READS,
            ColType.GENETIC_SEQUENCE,
            ColType.GENETIC_PROFILE,
            ColType.GENETIC_DISTANCE,
        }
    )
    ORGANIZATION = frozenset({ColType.ORGANIZATION})
    OTHER = frozenset({ColType.OTHER})
    HAS_CONCEPT_SET = frozenset(
        {
            ColType.NOMINAL,
            ColType.ORDINAL,
            ColType.INTERVAL,
            ColType.REGULAR_LANGUAGE,
            ColType.CONTEXT_FREE_GRAMMAR_JSON,
            ColType.CONTEXT_FREE_GRAMMAR_XML,
        }
    )
    HAS_REGION_SET = frozenset({ColType.GEO_REGION})
    HAS_GENETIC_DISTANCE_PROTOCOL = frozenset({ColType.GENETIC_DISTANCE})


class CaseColDataRule(Enum):
    MISSING = "MISSING"
    INVALID = "INVALID"
    UNAUTHORIZED = "UNAUTHORIZED"
    CONFLICT = "CONFLICT"
    DERIVED = "DERIVED"



# pylint: disable=useless-import-alias
from gen_epix.omopdb.domain.model.omop.base import DataLineageMixin as DataLineageMixin
from gen_epix.omopdb.domain.model.omop.non_persistable import Subject as Subject
from gen_epix.omopdb.domain.model.omop.omop import CareSite as CareSite
from gen_epix.omopdb.domain.model.omop.omop import CdmSource as CdmSource
from gen_epix.omopdb.domain.model.omop.omop import Cohort as Cohort
from gen_epix.omopdb.domain.model.omop.omop import CohortDefinition as CohortDefinition
from gen_epix.omopdb.domain.model.omop.omop import Concept as Concept
from gen_epix.omopdb.domain.model.omop.omop import ConceptAncestor as ConceptAncestor
from gen_epix.omopdb.domain.model.omop.omop import ConceptClass as ConceptClass
from gen_epix.omopdb.domain.model.omop.omop import (
    ConceptRelationship as ConceptRelationship,
)
from gen_epix.omopdb.domain.model.omop.omop import ConceptSynonym as ConceptSynonym
from gen_epix.omopdb.domain.model.omop.omop import ConditionEra as ConditionEra
from gen_epix.omopdb.domain.model.omop.omop import (
    ConditionOccurrence as ConditionOccurrence,
)
from gen_epix.omopdb.domain.model.omop.omop import Cost as Cost
from gen_epix.omopdb.domain.model.omop.omop import DeviceExposure as DeviceExposure
from gen_epix.omopdb.domain.model.omop.omop import Domain as Domain
from gen_epix.omopdb.domain.model.omop.omop import DoseEra as DoseEra
from gen_epix.omopdb.domain.model.omop.omop import DrugEra as DrugEra
from gen_epix.omopdb.domain.model.omop.omop import DrugExposure as DrugExposure
from gen_epix.omopdb.domain.model.omop.omop import DrugStrength as DrugStrength
from gen_epix.omopdb.domain.model.omop.omop import FactRelationship as FactRelationship
from gen_epix.omopdb.domain.model.omop.omop import Location as Location
from gen_epix.omopdb.domain.model.omop.omop import LocationHistory as LocationHistory
from gen_epix.omopdb.domain.model.omop.omop import Measurement as Measurement
from gen_epix.omopdb.domain.model.omop.omop import (
    MeasurementRelation as MeasurementRelation,
)
from gen_epix.omopdb.domain.model.omop.omop import Metadata as Metadata
from gen_epix.omopdb.domain.model.omop.omop import Note as Note
from gen_epix.omopdb.domain.model.omop.omop import NoteNlp as NoteNlp
from gen_epix.omopdb.domain.model.omop.omop import Observation as Observation
from gen_epix.omopdb.domain.model.omop.omop import (
    ObservationPeriod as ObservationPeriod,
)
from gen_epix.omopdb.domain.model.omop.omop import PayerPlanPeriod as PayerPlanPeriod
from gen_epix.omopdb.domain.model.omop.omop import Person as Person
from gen_epix.omopdb.domain.model.omop.omop import (
    ProcedureOccurrence as ProcedureOccurrence,
)
from gen_epix.omopdb.domain.model.omop.omop import Provider as Provider
from gen_epix.omopdb.domain.model.omop.omop import Relationship as Relationship
from gen_epix.omopdb.domain.model.omop.omop import (
    SourceToConceptMap as SourceToConceptMap,
)
from gen_epix.omopdb.domain.model.omop.omop import Specimen as Specimen
from gen_epix.omopdb.domain.model.omop.omop import SurveyConduct as SurveyConduct
from gen_epix.omopdb.domain.model.omop.omop import VisitDetail as VisitDetail
from gen_epix.omopdb.domain.model.omop.omop import VisitOccurrence as VisitOccurrence
from gen_epix.omopdb.domain.model.omop.omop import Vocabulary as Vocabulary
